import os
import hydra
from hydra.utils import to_absolute_path
import numpy as np
import time
from lensless.utils.plot import plot_image
from lensless.utils.io import save_image
from lensless.utils.image import resize, gamma_correction
import matplotlib.pyplot as plt
from lensless import FISTA, ADMM
from lensless.utils.io import load_image, load_psf
import omegaconf
from lensless.hardware.trainable_mask import AdafruitLCD


@hydra.main(version_base=None, config_path="../../configs", config_name="demo")
def demo(config):

    start_time = time.time()

    if config.save:
        if config.output is not None:
            # make sure output directory exists
            os.makedirs(config.output, exist_ok=True)
            save = config.output
        else:
            save = os.getcwd()
    else:
        save = False

    # prepare data
    # -- load raw data
    localfile = f"{config.capture.raw_data_fn}.png"
    if save:
        localfile = os.path.join(save, localfile)
    print("\nLoading picture...")
    if config.capture.bayer:
        img = load_image(
            localfile,
            verbose=False,
            bayer=False if config.capture.rgb else True,
            blue_gain=config.camera.blue_gain,
            red_gain=config.camera.red_gain,
            nbits_out=config.capture.nbits_out,
        )
    else:
        img = load_image(localfile, verbose=True, bayer=False)

    data = np.array(img, dtype=config.recon.dtype)
    if len(data.shape) == 3:
        data = data[np.newaxis, :, :, :]
    elif len(data.shape) == 2:
        data = data[np.newaxis, :, :, np.newaxis]

    # -- PSF
    flipud = False
    if isinstance(config.camera.psf, omegaconf.dictconfig.DictConfig):
        import torch

        np.random.seed(config.camera.psf.seed % (2**32 - 1))
        mask_vals = np.random.uniform(0, 1, config.camera.psf.mask_shape)
        mask_vals_torch = torch.from_numpy(mask_vals.astype(np.float32))
        flipud = config.camera.psf.flipud
        mask = AdafruitLCD(
            initial_vals=mask_vals_torch,
            sensor=config.capture.sensor,
            slm=config.camera.psf.device,
            downsample=config.recon.downsample,
            flipud=flipud,
        )
        psf = mask.get_psf().detach().numpy()
        bg = np.zeros(psf.shape[-1])

    elif config.camera.background is not None:
        psf = load_psf(
            to_absolute_path(config.camera.psf),
            downsample=config.recon.downsample,
            return_float=True,
            return_bg=False,
            dtype=config.recon.dtype,
        )
        bg = np.load(to_absolute_path(config.camera.background))

    else:
        psf, bg = load_psf(
            to_absolute_path(config.camera.psf),
            downsample=config.recon.downsample,
            return_float=True,
            return_bg=True,
            dtype=config.recon.dtype,
        )
    psf = np.array(psf, dtype=config.recon.dtype)
    if config.plot:
        ax = plot_image(psf[0], gamma=config.recon.gamma)
        ax.set_title("PSF")
        if save:
            plt.savefig(os.path.join(save, "psf.png"))

    # -- prepare data
    if data.min() > 0:
        data -= bg
    data = np.clip(data, a_min=0, a_max=data.max())

    if data.shape != psf.shape:
        # in DiffuserCam dataset, images are already reshaped
        data = resize(data, shape=psf.shape)
    data /= np.linalg.norm(data.ravel())
    data = np.array(data, dtype=config.recon.dtype)

    if config.recon.use_torch:
        import torch

        if config.recon.dtype == "float32":
            torch_dtype = torch.float32
        elif config.recon.dtype == "float64":
            torch_dtype = torch.float64
        else:
            raise ValueError("dtype must be float32 or float64")

        psf = torch.from_numpy(psf).type(torch_dtype).to(config.recon.torch_device)
        data = torch.from_numpy(data).type(torch_dtype).to(config.recon.torch_device)
        if flipud:
            data = torch.rot90(data, dims=(-3, -2), k=2)
    else:
        if flipud:
            data = np.rot90(data, k=2, axes=(-3, -2))

    print(f"Setup time : {time.time() - start_time} s")

    # -- apply algo
    start_time = time.time()

    if config.recon.algo == "fista":
        algo_params = config.recon.fista
        recon = FISTA(
            psf,
            **algo_params,
        )
    elif config.recon.algo == "admm":
        algo_params = config.recon.admm
        recon = ADMM(
            psf,
            **algo_params,
        )
    elif config.recon.algo == "unrolled":
        assert config.recon.use_torch, "Unrolled ADMM only available with torch"
        from lensless import UnrolledADMM

        algo_params = config.recon.unrolled_admm
        recon = UnrolledADMM(
            psf,
            **algo_params,
        )
        print("Loading checkpoint from : ", algo_params.checkpoint_fp)
        assert os.path.exists(algo_params.checkpoint_fp), "Checkpoint does not exist"
        recon.load_state_dict(
            torch.load(algo_params.checkpoint_fp, map_location=config.recon.torch_device)
        )
    else:
        raise ValueError(f"Unsupported algorithm: {config.recon.algo}")

    print("Applying : ", config.recon.algo)

    if config.recon.use_torch:
        with torch.no_grad():
            recon.set_data(data)
            res = recon.apply(
                gamma=config.recon.gamma,
                save=save,
                plot=config.plot,
                disp_iter=algo_params["disp_iter"],
            )
    else:
        recon.set_data(data)
        res = recon.apply(
            gamma=config.recon.gamma,
            save=save,
            plot=config.plot,
            disp_iter=algo_params["disp_iter"],
        )
    print(f"Processing time : {time.time() - start_time} s")

    if config.plot:
        final_image = res[0]
    else:
        final_image = res

    # save final image ax
    if save:
        # take first depth
        final_image = final_image[0]
        if config.recon.use_torch:
            img = final_image.cpu().numpy()
        else:
            img = final_image

        if config.postproc.crop_hor is not None:
            img = img[
                :,
                int(config.postproc.crop_hor[0] * img.shape[1]) : int(
                    config.postproc.crop_hor[1] * img.shape[1]
                ),
            ]
        if config.postproc.crop_vert is not None:
            img = img[
                int(config.postproc.crop_vert[0] * img.shape[0]) : int(
                    config.postproc.crop_vert[1] * img.shape[0]
                ),
                :,
            ]

        img_norm = img / img.max()
        if config.recon.gamma:
            img_norm = gamma_correction(img_norm, gamma=config.recon.gamma)

        output_fp = os.path.join(save, "reconstructed.png")
        save_image(img_norm, output_fp)

    if config.plot:
        plt.show()

    return save


if __name__ == "__main__":
    demo()



# #############################################################################
# benchmark_recon.py
# =========
# Authors :
# Yohann PERRON
# Eric BEZZAM [ebezzam@gmail.com]
# #############################################################################

"""
Benchmark reconstruction algorithms
==============
This script benchmarks reconstruction algorithms on the DiffuserCam test dataset.
The algorithm benchmarked and the number of iterations can be set in the config file : benchmark.yaml.
For unrolled algorithms, the results of the unrolled training (json file) are loaded from the benchmark/results folder.
"""

import hydra
from hydra.utils import get_original_cwd

import time
import numpy as np
import glob
import json
import os
from waveprop.noise import add_shot_noise
import pathlib as plib
from lensless.eval.benchmark import benchmark
import matplotlib.pyplot as plt
from lensless import ADMM, FISTA, GradientDescent, NesterovGradientDescent
from lensless.utils.dataset import DiffuserCamTestDataset, DigiCamCelebA, HFDataset
from lensless.utils.io import save_image
from lensless.utils.image import gamma_correction
from lensless.recon.model_dict import download_model, load_model

import torch
from torch.utils.data import Subset


@hydra.main(version_base=None, config_path="../../configs/benchmark", config_name="defaults")
def benchmark_recon(config):

    # set seed
    torch.manual_seed(config.seed)
    np.random.seed(config.seed)
    generator = torch.Generator().manual_seed(config.seed)

    downsample = config.downsample
    n_files = config.n_files
    n_iter_range = config.n_iter_range
    pnp = config.pnp  # parameterize and perturn parameters

    # check if GPU is available
    if torch.cuda.is_available() and config.device[:4] == "cuda":
        device = config.device
    else:
        device = "cpu"

    # Benchmark dataset
    crop = None
    dataset = config.dataset
    if dataset == "DiffuserCam":
        benchmark_dataset = DiffuserCamTestDataset(n_files=n_files, downsample=downsample)
        psf = benchmark_dataset.psf.to(device)

    elif dataset == "DigiCamCelebA":

        dataset = DigiCamCelebA(
            data_dir=os.path.join(get_original_cwd(), config.files.dataset),
            celeba_root=config.files.celeba_root,
            psf_path=os.path.join(get_original_cwd(), config.files.psf),
            downsample=config.files.downsample,
            vertical_shift=config.files.vertical_shift,
            horizontal_shift=config.files.horizontal_shift,
            simulation_config=config.simulation,
            crop=config.files.crop,
        )
        dataset.psf = dataset.psf.to(device)
        psf = dataset.psf
        crop = dataset.crop

        if config.n_files is not None:
            dataset = Subset(dataset, np.arange(config.n_files))
            dataset.psf = dataset.dataset.psf

        # train-test split
        train_size = int((1 - config.files.test_size) * len(dataset))
        test_size = len(dataset) - train_size
        _, benchmark_dataset = torch.utils.data.random_split(
            dataset, [train_size, test_size], generator=generator
        )
    elif dataset == "HFDataset":

        split_test = "test"
        if config.huggingface.split_seed is not None:
            from datasets import load_dataset, concatenate_datasets

            seed = config.huggingface.split_seed
            generator = torch.Generator().manual_seed(seed)

            # - combine train and test into single dataset
            train_split = "train"
            test_split = "test"
            if config.n_files is not None:
                train_split = f"train[:{config.n_files}]"
                test_split = f"test[:{config.n_files}]"
            train_dataset = load_dataset(
                config.huggingface.repo, split=train_split, cache_dir=config.huggingface.cache_dir
            )
            test_dataset = load_dataset(
                config.huggingface.repo, split=test_split, cache_dir=config.huggingface.cache_dir
            )
            dataset = concatenate_datasets([test_dataset, train_dataset])

            # - split into train and test
            train_size = int((1 - config.files.test_size) * len(dataset))
            test_size = len(dataset) - train_size
            _, split_test = torch.utils.data.random_split(
                dataset, [train_size, test_size], generator=generator
            )

        benchmark_dataset = HFDataset(
            huggingface_repo=config.huggingface.repo,
            cache_dir=config.huggingface.cache_dir,
            psf=config.huggingface.psf,
            n_files=n_files,
            split=split_test,
            display_res=config.huggingface.image_res,
            rotate=config.huggingface.rotate,
            flipud=config.huggingface.flipud,
            flip_lensed=config.huggingface.flip_lensed,
            downsample=config.huggingface.downsample,
            downsample_lensed=config.huggingface.downsample_lensed,
            alignment=config.huggingface.alignment,
            simulation_config=config.simulation,
            single_channel_psf=config.huggingface.single_channel_psf,
        )
        if benchmark_dataset.multimask:
            # get first PSF for initialization
            first_psf_key = list(benchmark_dataset.psf.keys())[0]
            psf = benchmark_dataset.psf[first_psf_key].to(device)
        else:
            psf = benchmark_dataset.psf.to(device)
    else:
        raise ValueError(f"Dataset {dataset} not supported")

    print(f"Number of files : {len(benchmark_dataset)}")
    print(f"Data shape :  {benchmark_dataset[0][0].shape}")

    model_list = []  # list of algoritms to benchmark
    for algo in config.algorithms:
        if algo == "ADMM":
            model_list.append(
                (
                    "ADMM",
                    ADMM(
                        psf,
                        mu1=config.admm.mu1,
                        mu2=config.admm.mu2,
                        mu3=config.admm.mu3,
                        tau=config.admm.tau,
                    ),
                )
            )
        if algo == "ADMM_Monakhova2019":
            model_list.append(
                ("ADMM_Monakhova2019", ADMM(psf, mu1=1e-4, mu2=1e-4, mu3=1e-4, tau=2e-3))
            )
        if algo == "ADMM_PnP":
            model_list.append(
                (
                    "ADMM_PnP",
                    ADMM(
                        psf,
                        mu1=config.admm.mu1,
                        mu2=config.admm.mu2,
                        mu3=config.admm.mu3,
                        tau=config.admm.tau,
                        denoiser={"network": "DruNet", "noise_level": 30, "use_dual": False},
                    ),
                )
            )
        if algo == "FISTA_PnP":
            model_list.append(
                (
                    "FISTA_PnP",
                    FISTA(
                        psf,
                        tk=config.fista.tk,
                        denoiser={"network": "DruNet", "noise_level": 30},
                    ),
                )
            )
        if algo == "FISTA":
            model_list.append(("FISTA", FISTA(psf, tk=config.fista.tk)))
        if algo == "GradientDescent":
            model_list.append(("GradientDescent", GradientDescent(psf)))
        if algo == "NesterovGradientDescent":
            model_list.append(
                (
                    "NesterovGradientDescent",
                    NesterovGradientDescent(psf, p=config.nesterov.p, mu=config.nesterov.mu),
                )
            )
        if "hf" in algo:
            param = algo.split(":")
            assert (
                len(param) == 4
            ), "hf model requires following format: hf:camera:dataset:model_name"
            camera = param[1]
            dataset = param[2]
            model_name = param[3]
            algo_config = config.get(algo)
            if algo_config is not None:
                skip_pre = algo_config.get("skip_pre", False)
                skip_post = algo_config.get("skip_post", False)
            else:
                skip_pre = False
                skip_post = False

            model_path = download_model(camera=camera, dataset=dataset, model=model_name)
            model_list.append((algo, model_path))

    results = {}
    output_dir = None

    # save PSF
    psf_np = psf.cpu().numpy()[0]
    psf_np = psf_np / np.max(psf_np)
    psf_np = gamma_correction(psf_np, gamma=config.gamma_psf)
    save_image(psf_np, fp="psf.png")

    # save ground truth and lensless images
    if config.save_idx is not None:

        assert np.max(config.save_idx) < len(
            benchmark_dataset
        ), "save_idx values must be smaller than dataset size"

        os.mkdir("GROUND_TRUTH")
        os.mkdir("LENSLESS")
        if benchmark_dataset.measured_bg:
            os.mkdir("BACKGROUND")
        for idx in config.save_idx:
            data = benchmark_dataset[idx]
            lensless = data[0]
            if config.snr is not None:
                # example of target SNR
                lensless = add_shot_noise(lensless, float(config.snr))
            ground_truth = data[1]
            ground_truth_np = ground_truth.cpu().numpy()[0]
            lensless_np = lensless.cpu().numpy()[0]

            if crop is not None:
                ground_truth_np = ground_truth_np[
                    crop["vertical"][0] : crop["vertical"][1],
                    crop["horizontal"][0] : crop["horizontal"][1],
                ]

            save_image(
                ground_truth_np,
                fp=os.path.join("GROUND_TRUTH", f"{idx}.png"),
            )
            save_image(
                lensless_np,
                fp=os.path.join("LENSLESS", f"{idx}.png"),
            )
            if benchmark_dataset.measured_bg:
                background = data[-1]
                background_np = background.cpu().numpy()[0]
                save_image(
                    background_np,
                    fp=os.path.join("BACKGROUND", f"{idx}_background.png"),
                )

    # benchmark each model for different number of iteration and append result to results
    # -- batchsize has to equal 1 as baseline models don't support batch processing
    start_time = time.time()
    for model_name, model in model_list:

        if config.save_idx is not None:
            # make directory for outputs
            os.mkdir(model_name)

        results[model_name] = dict()

        if "hf" in model_name:
            # trained algorithm
            print(f"Running benchmark for {model_name}")

            # -- load model
            model_obj = load_model(
                model,  # model path
                psf,
                device,
                skip_pre=skip_pre,
                skip_post=skip_post,
                return_intermediate=config.save_intermediate,
            )
            model_obj.eval()

            if pnp is not None:
                pnp["model_path"] = model
                print(f"Using parameterize and perturb (P&P) with {pnp} parameters...")

            result = benchmark(
                model_obj,
                benchmark_dataset,
                batchsize=config.batchsize,
                save_idx=config.save_idx,
                save_intermediate=config.save_intermediate,
                output_dir=model_name,
                crop=crop,
                snr=config.snr,
                pnp=pnp,
                swap_channels=config.swap_channels,
            )
            results[model_name] = result

            # -- save results as easy to read JSON
            results_path = "results.json"
            with open(results_path, "w") as f:
                json.dump(results, f, indent=4)

        else:
            # iterative algorithm
            if pnp is not None:
                print(f"Parameterize and perturb (P&P) not used for {model_name}...")

            for n_iter in n_iter_range:

                print(f"Running benchmark for {model_name} with {n_iter} iterations")

                if config.save_idx is not None:
                    output_dir = os.path.join(model_name, str(n_iter))
                    os.mkdir(output_dir)

                result = benchmark(
                    model,
                    benchmark_dataset,
                    batchsize=1,
                    n_iter=n_iter,
                    save_idx=config.save_idx,
                    save_intermediate=config.save_intermediate,
                    output_dir=output_dir,
                    crop=crop,
                    use_background=config.huggingface.use_background,
                    snr=config.snr,
                    swap_channels=config.swap_channels,
                )
                results[model_name][int(n_iter)] = result

                # -- save results as easy to read JSON
                results_path = "results.json"
                with open(results_path, "w") as f:
                    json.dump(results, f, indent=4)
    proc_time = (time.time() - start_time) / 60
    print(f"Total processing time: {proc_time:.2f} min")

    # create folder to load results from trained algorithms
    result_dir = os.path.join(get_original_cwd(), "benchmark", "trained_results")
    if not os.path.isdir(result_dir):
        os.mkdir(result_dir)

    # try to load json files with results form unrolled training
    files = glob.glob(os.path.join(result_dir, "*.json"))
    unrolled_results = {}
    for file in files:
        model_name = plib.Path(file).stem
        unrolled_results[model_name] = {}
        with open(file, "r") as f:
            result = json.load(f)

            # get result for each metric
            for metric in result.keys():
                # if list take last value (last epoch)
                if isinstance(result[metric], list):
                    unrolled_results[model_name][metric] = result[metric][-1]
                else:
                    unrolled_results[model_name][metric] = result[metric]

    # Baseline results
    baseline_label = config.baseline
    baseline_results = None
    if dataset == "DiffuserCam":
        # (Monakhova et al. 2019, https://arxiv.org/abs/1908.11502)
        # -- ADMM (100)
        if baseline_label == "MONAKHOVA 100iter":
            baseline_results = {
                "MSE": 0.0622,
                "LPIPS_Alex": 0.5711,
                "ReconstructionError": 13.62,
            }
        # -- ADMM (5)
        elif baseline_label == "MONAKHOVA 5iter":
            baseline_results = {
                "MSE": 0.1041,
                "LPIPS_Alex": 0.6309,
                "ReconstructionError": 11.32,
            }
        # -- Le-ADMM (Unrolled 5)
        elif baseline_label == "MONAKHOVA Unrolled 5iter":
            baseline_results = {
                "MSE": 0.0618,
                "LPIPS_Alex": 0.4434,
                "ReconstructionError": 13.70,
            }
        # -- Le-ADMM-U (Unrolled 5 + UNet post-denoiser)
        elif baseline_label == "MONAKHOVA Unrolled 5iter + UNet":
            baseline_results = {
                "MSE": 0.0074,
                "LPIPS_Alex": 0.1904,
                "ReconstructionError": 22.14,
            }
        else:
            raise ValueError(f"Baseline {baseline_label} not supported")

    # for each metrics plot the results comparing each model
    metrics_to_plot = ["SSIM", "PSNR", "MSE", "LPIPS_Vgg", "LPIPS_Alex", "ReconstructionError"]

    if "hf" in model_name:
        available_metrics = list(results[model_name].keys())
    else:
        available_metrics = list(results[model_name][n_iter_range[0]].keys())
    metrics_to_plot = [metric for metric in metrics_to_plot if metric in available_metrics]
    # print metrics being skipped
    skipped_metrics = [metric for metric in metrics_to_plot if metric not in available_metrics]
    if len(skipped_metrics) > 0:
        print(f"Metrics {skipped_metrics} not available and will be skipped")
    for metric in metrics_to_plot:
        plt.figure()
        # plot benchmarked algorithm
        for model_name in results.keys():
            if "hf" in model_name:
                # doesn't change over number of iterations as assumed fixed unrolled
                continue
            plt.plot(
                n_iter_range,
                [results[model_name][n_iter][metric] for n_iter in n_iter_range],
                label=model_name,
            )
        # plot baseline as horizontal dotted line
        if baseline_results is not None:
            if metric in baseline_results.keys():
                plt.hlines(
                    baseline_results[metric],
                    0,
                    max(n_iter_range),
                    linestyles="dashed",
                    label=baseline_label,
                    color="orange",
                )

        # plot unrolled algorithms results
        color_list = ["red", "green", "blue", "orange", "purple"]
        algorithm_colors = {}
        for model_name in unrolled_results.keys():
            # use algorithm name if defined, else use file name
            if "algorithm" in unrolled_results[model_name].keys():
                plot_name = unrolled_results[model_name]["algorithm"]
            else:
                plot_name = model_name

            # set color depending on plot name using same color for same algorithm
            first = False
            if plot_name not in algorithm_colors.keys():
                algorithm_colors[plot_name] = color_list.pop()
                first = True
            color = algorithm_colors[plot_name]

            # check if metric is defined
            if metric not in unrolled_results[model_name].keys():
                continue
            # if n_iter is undefined, plot as horizontal line
            if "n_iter" not in unrolled_results[model_name].keys():
                plt.hlines(
                    unrolled_results[model_name][metric],
                    0,
                    n_iter_range[-1],
                    label=plot_name,
                    linestyles="dashed",
                    colors=color,
                )
            else:
                # plot as point
                if first:
                    plt.plot(
                        unrolled_results[model_name]["n_iter"],
                        unrolled_results[model_name][metric],
                        label=plot_name,
                        marker="o",
                        color=color,
                    )
                else:
                    plt.plot(
                        unrolled_results[model_name]["n_iter"],
                        unrolled_results[model_name][metric],
                        marker="o",
                        color=color,
                    )
        plt.xlabel("Number of iterations", fontsize="12")
        plt.ylabel(metric, fontsize="12")
        plt.legend(fontsize="12")
        plt.grid()
        plt.savefig(f"{metric}.png")


if __name__ == "__main__":
    benchmark_recon()



from diffusercam_gd import grad_descent, GradientDescentUpdate
from lensless.recon.gd import FISTA
from lensless.recon.apgd import APGD, APGDPriors
from lensless.utils.io import load_data
import time
import pathlib as plib


psf_fp = "data/psf/tape_rgb.png"
data_fp = "data/raw_data/thumbs_up_rgb.png"

n_iter = 300
downsample = 4
gray = True
dtype = "float32"
n_trials = 3

psf, data = load_data(
    psf_fp=psf_fp,
    data_fp=data_fp,
    downsample=downsample,
    plot=False,
    gray=gray,
    dtype=dtype,
)

save = "profile_gradient_descent"
save = plib.Path(__file__).parent / save
save.mkdir(exist_ok=True)

""" LenslessPiCam """
recon = FISTA(psf, dtype=dtype)
recon.set_data(data)
# res = recon.apply(n_iter=n_iter, save=save, disp_iter=None)
# recon.reset()
total_time = 0
for _ in range(n_trials):
    start_time = time.time()
    res = recon.apply(n_iter=n_iter, disp_iter=None)
    total_time += time.time() - start_time
    recon.reset()
print(f"LenslessPiCam (avg) : {total_time / n_trials} s")

# -- using Pycsou, complex conv
recon = APGD(
    psf,
    max_iter=n_iter,
    acceleration="BT",
    diff_penalty=None,
    prox_penalty=APGDPriors.NONNEG,
    realconv=False,
)
recon.set_data(data)
res = recon.apply(n_iter=n_iter, save=save, disp_iter=None)
recon.reset()
total_time = 0
for _ in range(n_trials):
    start_time = time.time()
    res = recon.apply(n_iter=n_iter, disp_iter=None)
    total_time += time.time() - start_time
    recon.reset()
print(f"LenslessPiCam (Pycsou, complex) : {total_time / n_trials} s")

# -- using Pycsou, real conv
recon = APGD(
    psf,
    max_iter=n_iter,
    acceleration="BT",
    diff_penalty=None,
    prox_penalty=APGDPriors.NONNEG,
    realconv=True,
)
recon.set_data(data)
res = recon.apply(n_iter=n_iter, save=save, disp_iter=None)
recon.reset()
total_time = 0
for _ in range(n_trials):
    start_time = time.time()
    res = recon.apply(n_iter=n_iter, disp_iter=None)
    total_time += time.time() - start_time
    recon.reset()
print(f"LenslessPiCam (Pycsou, real) : {total_time / n_trials} s")

""" DiffuserCam"""
method = GradientDescentUpdate.FISTA
grad_descent(psf, data, n_iter=n_iter, update_method=method, disp_iter=None, save=save)
start_time = time.time()
for _ in range(n_trials):
    grad_descent(psf, data, n_iter=n_iter, update_method=method, disp_iter=None)
print(f"DiffuserCam : {(time.time() - start_time) / n_trials} s")

""" PyTorch CPU """
psf, data = load_data(
    psf_fp=psf_fp,
    data_fp=data_fp,
    downsample=downsample,
    plot=False,
    gray=gray,
    dtype=dtype,
    use_torch=True,
)

recon = FISTA(psf, dtype=dtype)
recon.set_data(data)
res = recon.apply(n_iter=n_iter, save=save, disp_iter=None, plot=False)
recon.reset()
total_time = 0
for _ in range(n_trials):
    start_time = time.time()
    res = recon.apply(n_iter=n_iter, disp_iter=None, plot=False)
    total_time += time.time() - start_time
    recon.reset()
print(f"LenslessPiCam, PyTorch CPU (avg) : {total_time / n_trials} s")

""" PyTorch GPU """
psf, data = load_data(
    psf_fp=psf_fp,
    data_fp=data_fp,
    downsample=downsample,
    plot=False,
    gray=gray,
    dtype=dtype,
    use_torch=True,
    torch_device="cuda",
)

recon = FISTA(psf, dtype=dtype)
recon.set_data(data)
res = recon.apply(n_iter=n_iter, save=save, disp_iter=None, plot=False)
recon.reset()
total_time = 0
for _ in range(n_trials):
    start_time = time.time()
    res = recon.apply(n_iter=n_iter, disp_iter=None, plot=False)
    total_time += time.time() - start_time
    recon.reset()
print(f"LenslessPiCam, PyTorch GPU (avg) : {total_time / n_trials} s")



import hydra
import yaml
import torch
from lensless import ADMM
from lensless.utils.plot import plot_image
from lensless.utils.dataset import HFDataset
import os
from lensless.utils.io import save_image
import time
from lensless.recon.model_dict import download_model, load_model
from huggingface_hub import hf_hub_download
from lensless.utils.io import load_image


@hydra.main(version_base=None, config_path="../../configs/recon", config_name="digicam")
def apply_pretrained(config):
    save = config.save
    device = config.device
    n_trials = config.n_trials
    model_name = config.model

    # load config
    if model_name == "admm":
        # take config from unrolled ADMM for dataset
        model_path = download_model(camera="digicam", dataset="mirflickr_single_25k", model="U10")
        config_path = os.path.join(model_path, ".hydra", "config.yaml")
        with open(config_path, "r") as stream:
            model_config = yaml.safe_load(stream)

    else:
        model_path = download_model(camera="digicam", dataset=config.dataset, model=model_name)
        config_path = os.path.join(model_path, ".hydra", "config.yaml")
        with open(config_path, "r") as stream:
            model_config = yaml.safe_load(stream)

    # load data
    # TODO try with multi-mask, should load single mask dataset...
    test_set = HFDataset(
        huggingface_repo=model_config["files"]["dataset"],
        psf=(
            model_config["files"]["huggingface_psf"]
            if "huggingface_psf" in model_config["files"]
            else None
        ),
        split="test",
        display_res=model_config["files"]["image_res"],
        rotate=model_config["files"]["rotate"],
        downsample=model_config["files"]["downsample"],
        alignment=model_config["alignment"],
        save_psf=model_config["files"]["save_psf"],
        simulation_config=model_config["simulation"],
        force_rgb=model_config["files"].get("force_rgb", False),
        cache_dir=config.cache_dir,
    )
    psf = test_set.psf.to(device)
    print("Test set size: ", len(test_set))

    if config.fn is not None:
        raw_data_fp = hf_hub_download(
            repo_id=model_config["files"]["dataset"], filename=config.fn, repo_type="dataset"
        )
        lensless = load_image(
            fp=raw_data_fp,
            return_float=True,
            as_4d=True,
        )
        lensless = torch.from_numpy(lensless).to(psf)
        if config.rotate:
            lensless = torch.rot90(lensless, dims=(-3, -2), k=2)

        lensed = None
        idx = os.path.basename(raw_data_fp).split("_")[0]

    else:
        idx = config.idx
        lensless, lensed = test_set[idx]
        lensless = lensless.to(device)

    # load model
    if model_name == "admm":
        recon = ADMM(psf, n_iter=config.n_iter)
    else:
        # load best model
        recon = load_model(model_path, psf, device)

    # print data shape
    print(f"Data shape :  {lensless.shape}")

    # apply reconstruction
    start_time = time.time()
    for _ in range(n_trials):
        with torch.no_grad():

            recon.set_data(lensless)
            res = recon.apply(
                disp_iter=-1,
                save=False,
                gamma=None,
                plot=False,
            )
    end_time = time.time()
    avg_time_ms = (end_time - start_time) / n_trials * 1000
    print(f"Avg inference [ms] : {avg_time_ms} ms")

    img = res[0].cpu().numpy().squeeze()

    plot_image(img)
    if lensed is not None:
        plot_image(lensed)

    if save:
        print(f"Saving images to {os.getcwd()}")

        if config.fn is not None:
            dim = config.alignment.dim
            top_left = config.alignment.top_left
            res_np = img[top_left[0] : top_left[0] + dim[0], top_left[1] : top_left[1] + dim[1]]
            bn = config.fn.split(".")[0]
            save_image(res_np, f"{model_name}_{bn}.png")
        else:
            alignment = test_set.alignment
            top_left = alignment["top_left"]
            height = alignment["height"]
            width = alignment["width"]
            res_np = img[top_left[0] : top_left[0] + height, top_left[1] : top_left[1] + width]
            save_image(res_np, f"{model_name}_{idx}.png")
        if lensed is not None:
            lensed_np = lensed[0].cpu().numpy()
            save_image(lensed_np, f"original_{idx}.png")
        save_image(lensless[0].cpu().numpy(), f"lensless_{idx}.png")


if __name__ == "__main__":
    apply_pretrained()



"""

This script shows:
1) how to simulate a dataset
2) apply a reconstruction algorithm
3) compute metrics

"""

import hydra
from hydra.utils import to_absolute_path
from lensless.utils.io import save_image
from lensless.utils.image import rgb2gray, resize
import numpy as np
from lensless import ADMM
from lensless.eval.metric import mse, psnr, ssim, lpips, LPIPS_MIN_DIM
from waveprop.simulation import FarFieldSimulator
import glob
import os
from tqdm import tqdm

from lensless.utils.io import load_image, load_psf


@hydra.main(version_base=None, config_path="../../configs/simulate", config_name="simulate_dataset")
def simulate(config):

    # set seed
    np.random.seed(config.seed)

    dataset = to_absolute_path(config.files.dataset)
    if not os.path.isdir(dataset):
        print(f"No dataset found at {dataset}")
        try:
            from torchvision.datasets.utils import download_and_extract_archive
        except ImportError:
            exit()
        msg = "Do you want to download the sample CelebA dataset (764KB)?"

        # default to yes if no input is given
        valid = input("%s (Y/n) " % msg).lower() != "n"
        if valid:
            url = "https://drive.switch.ch/index.php/s/Q5OdDQMwhucIlt8/download"
            filename = "celeb_mini.zip"
            download_and_extract_archive(
                url, os.path.dirname(dataset), filename=filename, remove_finished=True
            )

    psf_fp = to_absolute_path(config.files.psf)
    assert os.path.exists(psf_fp), f"PSF {psf_fp} does not exist."

    if config.save.bool:
        save_dir = to_absolute_path(config.save.dir)
        if not os.path.exists(save_dir):
            os.makedirs(save_dir)
            os.makedirs(os.path.join(save_dir, "sensor_plane"))
            os.makedirs(os.path.join(save_dir, "object_plane"))
            os.makedirs(os.path.join(save_dir, "reconstruction"))

    # load psf as numpy array
    print("\nPSF:")
    psf = load_psf(psf_fp, verbose=True, downsample=config.simulation.downsample)

    # remove depth dimension, as 3D not supported by FarFieldSimulator
    psf_sim = psf[0]
    if config.simulation.grayscale and len(psf_sim.shape) == 3:
        psf_sim = rgb2gray(psf_sim)
    if config.simulation.downsample > 1:
        print(f"Downsampled to {psf_sim.shape}.")

    # prepare simulator object
    simulator = FarFieldSimulator(psf=psf_sim, **config.simulation)

    # loop over files in dataset
    print("\nSimulating dataset...")
    files = glob.glob(os.path.join(dataset, f"*.{config.files.image_ext}"))
    if config.files.n_files is not None:
        files = files[: config.files.n_files]
    for fp in tqdm(files):

        # load image as numpy array
        image = load_image(fp)
        if config.simulation.grayscale and len(image.shape) == 3:
            image = rgb2gray(image)

        # simulate image
        image_plane, object_plane = simulator.propagate(image, return_object_plane=True)
        if config.save.bool:

            bn = os.path.basename(fp).split(".")[0] + ".png"

            # can serve as ground truth
            object_plane_fp = os.path.join(save_dir, "object_plane", bn)
            save_image(object_plane, object_plane_fp)  # use max range of 255

            # lensless image
            lensless_fp = os.path.join(save_dir, "sensor_plane", bn)
            save_image(image_plane, lensless_fp, max_val=config.simulation.max_val)

    # reconstruction
    if config.admm.bool:

        print("\nReconstructing lensless measurements...")

        output_dim = image_plane.shape
        if config.simulation.output_dim is not None:
            # best would be to incorporate downsampling in the reconstruction
            # for now downsample the PSF
            print("-- Resizing PSF to", config.simulation.output_dim, "for reconstruction.")
            psf = resize(psf, shape=config.simulation.output_dim)

        # -- initialize reconstruction object
        recon = ADMM(psf, **config.admm)

        # -- metrics
        mse_vals = []
        psnr_vals = []
        ssim_vals = []
        if not config.simulation.grayscale and np.min(output_dim[:2]) >= LPIPS_MIN_DIM:
            lpips_vals = []
        else:
            lpips_vals = None

        # -- loop over files in dataset
        files = glob.glob(os.path.join(save_dir, "sensor_plane", "*.png"))
        if config.files.n_files is not None:
            files = files[: config.files.n_files]

        for fp in tqdm(files):

            lensless = load_image(fp, as_4d=True)
            lensless = lensless / np.max(lensless)
            recon.set_data(lensless)
            res, _ = recon.apply(n_iter=config.admm.n_iter, disp_iter=config.admm.disp_iter)
            recovered = res[0]  # first depth

            if config.save.bool:
                bn = os.path.basename(fp).split(".")[0] + ".png"
                lensless_fp = os.path.join(save_dir, "reconstruction", bn)
                save_image(recovered, lensless_fp, max_val=config.simulation.max_val)

            # compute metrics
            object_plane_fp = os.path.join(save_dir, "object_plane", os.path.basename(fp))
            object_plane = load_image(object_plane_fp)

            if config.simulation.output_dim is not None:
                # best would be to incorporate downsampling in the reconstruction
                # for now downsample the PSF
                # print("-- Resizing object plane to", config.simulation.output_dim, "for metric.")
                object_plane = resize(object_plane, shape=config.simulation.output_dim)

            mse_vals.append(mse(object_plane, recovered))
            psnr_vals.append(psnr(object_plane, recovered))
            if config.simulation.grayscale:
                ssim_vals.append(ssim(object_plane, recovered, channel_axis=None))
            else:
                ssim_vals.append(ssim(object_plane, recovered))
            if lpips_vals is not None:
                lpips_vals.append(lpips(object_plane, recovered))

    print("\nMSE (avg)", np.mean(mse_vals))
    print("PSNR (avg)", np.mean(psnr_vals))
    print("SSIM (avg)", np.mean(ssim_vals))
    if lpips_vals is not None:
        print("LPIPS (avg)", np.mean(lpips_vals))


if __name__ == "__main__":
    simulate()



from lensless.recon.rfft_convolve import RealFFTConvolve2D
import numpy as np
import torch

psf_fp = "data/psf/tape_rgb.png"
data_fp = "data/raw_data/thumbs_up_rgb.png"
downsample = 8


# test if padding and cropping works
def test_pad_np():
    psf = np.random.rand(5, 47, 29, 3)
    convolver = RealFFTConvolve2D(psf, pad=True)
    data = np.random.rand(12, 5, 47, 29, 3)
    padded = convolver._pad(data)
    np.testing.assert_array_equal(padded.shape[1:], convolver._padded_shape)
    cropped = convolver._crop(padded)
    np.testing.assert_array_equal(cropped, data)


def test_pad_torch():
    psf = torch.rand(5, 47, 29, 3)
    convolver = RealFFTConvolve2D(psf, pad=True)
    data = torch.rand(12, 5, 47, 29, 3)
    padded = convolver._pad(data)
    np.testing.assert_array_equal(padded.shape[1:], convolver._padded_shape)
    cropped = convolver._crop(padded)
    assert torch.equal(cropped, data)


# test if convolution works identically on each color channel
def test_conv_np():
    psf = np.random.rand(5, 47, 29, 3)
    convolver = RealFFTConvolve2D(psf, pad=True)
    data = np.random.rand(12, 5, 47, 29, 3)
    all_convolved = convolver.convolve(data)
    part_convolved = convolver.convolve(data[:1, :1, :, :, :1])
    np.testing.assert_almost_equal(
        all_convolved[:1, :1, :, :, :1], part_convolved[:1, :1, :, :, :1], decimal=5
    )


def test_conv_torch():
    psf = torch.rand(5, 47, 29, 3)
    convolver = RealFFTConvolve2D(psf, pad=True)
    data = torch.rand(12, 1, 47, 29, 3)
    all_convolved = convolver.convolve(data)
    part_convolved = convolver.convolve(data[:1, :1, :, :, :])
    torch.testing.assert_close(all_convolved[:1, :1, :, :, :], part_convolved[:1, :1, :, :, :])


def test_conv_torch_2():
    psf = torch.rand(1, 47, 29, 3)
    convolver = RealFFTConvolve2D(psf, pad=True)
    data = torch.rand(12, 1, 47, 29, 3)
    all_convolved = convolver.convolve(data)
    part_convolved = convolver.convolve(data[:1, :1, :, :, :])
    print(all_convolved.shape, part_convolved.shape)
    torch.testing.assert_close(all_convolved[:1, :1, :, :, :], part_convolved[:1, :1, :, :, :])



"""
Fine-tune ViT on CelebA dataset measured with lensless camera.
Original tutorial: https://huggingface.co/blog/fine-tune-vit

First, set-up HuggingFace libraries:
```
pip install datasets transformers[torch] scikit-learn tensorboardX
```

Raw measurement datasets can be download from SwitchDrive.
This will be done by the script if the dataset is not found.
```
# 10K measurements (13.1 GB)
python scripts/classify/train_celeba_vit.py \
data.measured=data/celeba_adafruit_random_2mm_20230720_10K

# 1K measurements (1.2 GB)
python scripts/classify/train_celeba_vit.py \
data.measured=data/celeba_adafruit_random_2mm_20230720_1K
```

Note that the CelebA dataset also needs to be available locally!
It can be download here: https://mmlab.ie.cuhk.edu.hk/projects/CelebA.html

In order to classify on reconstructed outputs, the following
script needs to be run to create the dataset of reconstructed
images:
```
# reconstruct with ADMM
python scripts/recon/dataset.py algo=admm \
input.raw_data=path/to/raw/data
```

To classify on raw downsampled images, the same script can be
used, e.g. with the following command (`algo=null` for no reconstruction):
```
python scripts/recon/dataset.py algo=null \
input.raw_data=path/to/raw/data \
preprocess.data_dim=[48,64]
```

Other hyperparameters for classification can be found in
`configs/train_celeba_classifier.yaml`.

# TODO: update with Hugging Face dataset: https://huggingface.co/datasets/bezzam/DigiCam-CelebA-10K

"""

import warnings
from transformers import ViTImageProcessor, ViTForImageClassification
from transformers import TrainingArguments, Trainer, TrainerCallback
import numpy as np
import torch
import os
from hydra.utils import to_absolute_path
import glob
import hydra
import random
from datasets import load_metric
from PIL import Image
import pandas as pd
import time
import torchvision.transforms as transforms
import torchvision.datasets as dset
from datasets import Dataset
from copy import deepcopy
from torchvision.transforms import (
    CenterCrop,
    Compose,
    Normalize,
    RandomHorizontalFlip,
    RandomResizedCrop,
    Resize,
    ToTensor,
)


class CustomCallback(TrainerCallback):
    def __init__(self, trainer) -> None:
        super().__init__()
        self._trainer = trainer

    def on_epoch_end(self, args, state, control, **kwargs):
        if control.should_evaluate:
            control_copy = deepcopy(control)
            self._trainer.evaluate(
                eval_dataset=self._trainer.train_dataset, metric_key_prefix="train"
            )
            return control_copy

    def on_step_end(self, args, state, control, **kwargs):
        if control.should_evaluate:
            control_copy = deepcopy(control)
            self._trainer.evaluate(
                eval_dataset=self._trainer.train_dataset, metric_key_prefix="train"
            )
            return control_copy

    def on_train_end(self, args, state, control, **kwargs):
        if control.should_evaluate:
            control_copy = deepcopy(control)
            self._trainer.evaluate(
                eval_dataset=self._trainer.train_dataset, metric_key_prefix="train"
            )
            return control_copy


@hydra.main(version_base=None, config_path="../../configs", config_name="train_celeba_classifier")
def train_celeba_classifier(config):

    seed = config.seed
    torch.manual_seed(seed)
    random.seed(seed)
    np.random.seed(seed)

    # check how many measured files
    measured_dataset = to_absolute_path(config.data.measured)
    if not os.path.isdir(measured_dataset):
        print(f"No dataset found at {measured_dataset}")
        try:
            from torchvision.datasets.utils import download_and_extract_archive
        except ImportError:
            exit()
        msg = "Do you want to download the CelebA dataset measured with a random Adafruit LCD pattern (13.1 GB)?"

        # default to yes if no input is given
        valid = input("%s (Y/n) " % msg).lower() != "n"
        if valid:
            url = "https://drive.switch.ch/index.php/s/9NNGCJs3DoBDGlY/download"
            filename = "celeba_adafruit_random_2mm_20230720_10K.zip"
            download_and_extract_archive(
                url, os.path.dirname(measured_dataset), filename=filename, remove_finished=True
            )
    measured_files = sorted(glob.glob(os.path.join(measured_dataset, "*.png")))
    print(f"Found {len(measured_files)} files in {measured_dataset}")

    if config.data.n_files is not None:
        n_files = config.data.n_files
        measured_files = measured_files[: config.data.n_files]
        print(f"Using {len(measured_files)} files")
    n_files = len(measured_files)

    # create dataset split
    attr = config.data.attr
    ds = dset.CelebA(
        root=config.data.original,
        split="all",
        download=False,
        transform=transforms.ToTensor(),
    )
    label_idx = ds.attr_names.index(attr)
    labels = ds.attr[:, label_idx][:n_files]

    # make dataset with measured data and corresponding labels
    df = pd.DataFrame(
        {
            "labels": labels,
            "image_file_path": measured_files,
        }
    )
    ds = Dataset.from_pandas(df)
    ds = ds.class_encode_column("labels")

    # -- train / test split
    test_size = config.data.test_size
    ds = ds.train_test_split(
        test_size=test_size, stratify_by_column="labels", seed=seed, shuffle=True
    )

    # prepare dataset
    model_name_or_path = "google/vit-base-patch16-224-in21k"
    processor = ViTImageProcessor.from_pretrained(model_name_or_path)

    # -- processors for train and val
    image_mean, image_std = processor.image_mean, processor.image_std
    size = processor.size["height"]

    normalize = Normalize(mean=image_mean, std=image_std)
    # _train_transforms = Compose(
    #         [
    #             # RandomResizedCrop(
    #             #     size,
    #             #     scale=(0.9, 1.0),
    #             #     ratio=(0.9, 1.1),
    #             # ),
    #             Resize(size),
    #             CenterCrop(size),
    #             RandomHorizontalFlip(),
    #             ToTensor(),
    #             normalize,
    #         ]
    #     )
    _train_transforms = []
    if config.augmentation.random_resize_crop:
        _train_transforms.append(
            RandomResizedCrop(
                size,
                scale=(0.9, 1.0),
                ratio=(0.9, 1.1),
            )
        )
    _train_transforms += [Resize(size), CenterCrop(size)]
    if config.augmentation.horizontal_flip:
        if config.data.raw:
            warnings.warn("Horizontal flip is not supported for raw data, Skipping!")
        else:
            _train_transforms.append(RandomHorizontalFlip())
    _train_transforms += [ToTensor(), normalize]
    _train_transforms = Compose(_train_transforms)

    _val_transforms = Compose(
        [
            Resize(size),
            CenterCrop(size),
            ToTensor(),
            normalize,
        ]
    )

    def train_transforms(examples):
        # Take a list of PIL images and turn them to pixel values
        examples["pixel_values"] = [
            _train_transforms(Image.open(fp)) for fp in examples["image_file_path"]
        ]
        return examples

    def val_transforms(examples):
        # Take a list of PIL images and turn them to pixel values
        examples["pixel_values"] = [
            _val_transforms(Image.open(fp)) for fp in examples["image_file_path"]
        ]
        return examples

    # transform dataset
    ds["train"].set_transform(train_transforms)
    ds["test"].set_transform(val_transforms)

    # data collator
    def collate_fn(batch):
        return {
            "pixel_values": torch.stack([x["pixel_values"] for x in batch]),
            "labels": torch.tensor([x["labels"] for x in batch]),
        }

    # evaluation metric
    metric = load_metric("accuracy")

    def compute_metrics(p):
        return metric.compute(predictions=np.argmax(p.predictions, axis=1), references=p.label_ids)

    # load model
    if config.train.prev is not None:
        model_path = to_absolute_path(config.train.prev)
    else:
        model_path = model_name_or_path

    labels = ds["train"].features["labels"].names
    model = ViTForImageClassification.from_pretrained(
        model_path,
        num_labels=len(labels),
        id2label={str(i): c for i, c in enumerate(labels)},
        label2id={c: str(i) for i, c in enumerate(labels)},
        hidden_dropout_prob=config.train.dropout,
        attention_probs_dropout_prob=config.train.dropout,
    )

    # configure training
    output_dir = (
        config.data.output_dir + f"-{config.data.attr}" + os.path.basename(measured_dataset)
    )

    training_args = TrainingArguments(
        output_dir=output_dir,
        per_device_train_batch_size=config.train.batch_size,
        evaluation_strategy="steps",
        eval_steps=100,
        save_steps=100,
        num_train_epochs=config.train.n_epochs,
        fp16=True,
        logging_steps=10,
        learning_rate=config.train.learning_rate,
        save_total_limit=2,
        remove_unused_columns=False,  # important to keep False
        push_to_hub=False,
        report_to="tensorboard",
        load_best_model_at_end=True,
    )

    trainer = Trainer(
        model=model,
        args=training_args,
        data_collator=collate_fn,
        compute_metrics=compute_metrics,
        tokenizer=processor,
        train_dataset=ds["train"],
        eval_dataset=ds["test"],
    )
    trainer.add_callback(CustomCallback(trainer))  # add accuracy on train set

    # train
    hydra_output = os.getcwd()
    print("Results saved to : ", hydra_output)

    start_time = time.time()
    train_results = trainer.train()
    trainer.save_model()
    trainer.log_metrics("train", train_results.metrics)
    trainer.save_metrics("train", train_results.metrics)
    trainer.save_state()

    # evaluate
    metrics = trainer.evaluate(ds["train"])
    trainer.log_metrics("train", metrics)
    trainer.save_metrics("train", metrics)
    metrics = trainer.evaluate(ds["test"])
    trainer.log_metrics("eval", metrics)
    trainer.save_metrics("eval", metrics)

    #
    hydra_output = os.getcwd()
    print("Results saved to : ", hydra_output)
    print(f"Training took {time.time() - start_time} seconds")


if __name__ == "__main__":
    train_celeba_classifier()



# #############################################################################
# __init__.py
# ===========
# Authors :
# Eric BEZZAM [ebezzam@gmail.com]
# #############################################################################

"""
Lensless library
===========================

Package for designing, simulating, and imaging with lensless cameras.
"""


from .recon.recon import ReconstructionAlgorithm
from .recon.admm import ADMM
from .recon.gd import (
    GradientDescent,
    NesterovGradientDescent,
    FISTA,
    GradientDescentUpdate,
)
from .recon.tikhonov import CodedApertureReconstruction
from .hardware.sensor import VirtualSensor, SensorOptions

try:
    from .recon.trainable_recon import TrainableReconstructionAlgorithm
    from .recon.unrolled_admm import UnrolledADMM
    from .recon.unrolled_fista import UnrolledFISTA
    from .recon.trainable_inversion import TrainableInversion
    from .recon.multi_wiener import MultiWiener
    from .recon.sv_deconvnet import SVDeconvNet
except Exception:
    pass

try:
    from .recon.apgd import APGD, APGDPriors

    pycsou_available = True
except Exception:
    pycsou_available = False

from .version import __version__



# This script is used to export each depth of a .npy 3D reconstruction into single tiff files for visualisation.

import os
import sys
import numpy as np
import cv2

if len(sys.argv) < 2:
    print("Error : no filename provided. Aborting.")
    sys.exit()

filename = sys.argv[1]
if not filename.endswith(".npy"):
    print("Error : file is not a .npy file. Aborting")
    sys.exit()

out_path = os.path.splitext(filename)[0]  # removing the file extension from the path, if any

data = np.load(filename).astype(np.float32)
data_shape = data.shape
n_dim = len(data_shape)

print("\nInput shape :", data_shape)

if n_dim == 2:
    print("As the shape has length 2, it will be interpreted as a single-layer grayscale image.")
    grayscale = True
    single_depth = True

elif n_dim == 3:
    print(
        "As the shape has length 3, could either be a multi-layer grayscale image or a single-layer rgb image."
    )
    if data_shape[2] == 3:
        print(
            "As the third dimension of the data is 3, it will be interpreted as a single-layer rgb image"
            "with data corresponding to (width, height, color channel)."
        )
        grayscale = False
        single_depth = True
    else:
        print(
            "As the third dimension of the data is not, it will be interpreted as a multi-layer grayscale image "
            "with data corresponding to (depth, width, height)."
        )
        grayscale = True
        single_depth = False

elif n_dim == 4:
    print("As the shape has length 4, it will be interpreted as a multi-layer rgb image.")
    grayscale = False
    single_depth = False

else:
    print("Error : data shape has invalid length :", n_dim, ", but should be 2, 3, or 4")
    grayscale = None
    single_depth = None
    quit()

if single_depth:
    if grayscale:
        if cv2.imwrite(out_path + "-out.tiff", data):
            print("Data exported succesfully in the", out_path + "-out.tiff file.")
        else:
            print("Error while exporting data in the", out_path + "-out.tiff file.")
    else:
        if cv2.imwrite(
            out_path + "-out.tiff", cv2.cvtColor(data.astype(np.uint8), cv2.COLOR_RGB2BGR)
        ):
            print("Data exported succesfully in the", out_path + "-out.tiff file.")
        else:
            print("Error while exporting data in the", out_path + "-out.tiff file.")

else:
    print(
        "As the data has several depth layers, it will be stored in the",
        out_path + "-out directory.",
    )

    if os.path.exists(out_path + "-out/."):
        print("Directory already existing. The files inside will be replaced.")
    else:
        print("Directory not existing yet, it will be created.")
        os.mkdir(out_path + "-out/")

    for i in range(data_shape[0]):
        path = out_path + "-out/layer" + ("0" if i < 10 else "") + str(i) + ".tiff"
        if grayscale:
            if cv2.imwrite(path, data[i]):
                print("Data exported succesfully in the", path, "file.")
            else:
                print("Error while exporting data in the", path, "file.")
        else:
            if cv2.imwrite(path, cv2.cvtColor(data[i].astype(np.uint8), cv2.COLOR_RGB2BGR)):
                print("Data exported succesfully in the", path, "file.")
            else:
                print("Error while exporting data in the", path, "file.")



import hydra
import yaml
import torch
from lensless import ADMM
from lensless.utils.plot import plot_image
import os
from lensless.utils.io import save_image
import time
from lensless.recon.model_dict import download_model, load_model
from lensless.utils.dataset import get_dataset


@hydra.main(version_base=None, config_path="../../configs/recon", config_name="diffusercam")
def apply_pretrained(config):
    save = config.save
    device = config.device
    n_trials = config.n_trials
    model_name = config.model
    dataset_name = "mirflickr"

    # load config
    if model_name == "admm":
        # take config from any model just to get dataset hyperparameters
        model_path = download_model(camera="diffusercam", dataset=dataset_name, model="U20")
        config_path = os.path.join(model_path, ".hydra", "config.yaml")
        with open(config_path, "r") as stream:
            model_config = yaml.safe_load(stream)

    else:
        model_path = download_model(camera="diffusercam", dataset=dataset_name, model=model_name)
        config_path = os.path.join(model_path, ".hydra", "config.yaml")
        with open(config_path, "r") as stream:
            model_config = yaml.safe_load(stream)
    model_config["files"]["cache_dir"] = config.cache_dir

    # load data
    test_set = get_dataset(
        dataset_name="diffusercam_mirflickr",
        split="test",
        **model_config["files"],
    )
    psf = test_set.psf.to(device)
    print("Test set size: ", len(test_set))

    # reconstruction specified example
    idx = config.idx
    lensless, lensed = test_set[idx]
    lensless = lensless.to(device)

    # load model
    if model_name == "admm":
        recon = ADMM(psf, n_iter=config.n_iter)
    else:
        # load best model
        recon = load_model(model_path, psf, device)

    # print data shape
    print(f"Data shape :  {lensless.shape}")

    # apply reconstruction
    start_time = time.time()
    for _ in range(n_trials):
        with torch.no_grad():

            recon.set_data(lensless)
            res = recon.apply(
                disp_iter=-1,
                save=False,
                gamma=None,
                plot=False,
            )
    end_time = time.time()
    avg_time_ms = (end_time - start_time) / n_trials * 1000
    print(f"Avg inference [ms] : {avg_time_ms} ms")

    img = res[0].cpu().numpy().squeeze()

    plot_image(img)
    if lensed is not None:
        plot_image(lensed)

    if save:
        print(f"Saving images to {os.getcwd()}")
        save_image(img, f"{model_name}_{idx}.png")
        if lensed is not None:
            lensed_np = lensed[0].cpu().numpy()
            save_image(lensed_np, f"original_{idx}.png")
        save_image(lensless[0].cpu().numpy(), f"lensless_{idx}.png")


if __name__ == "__main__":
    apply_pretrained()



import sys
import numpy as np
from lensless.utils.image import resize3d

if len(sys.argv) < 2:
    print("Error : no filename provided. Aborting.")
    sys.exit()

filename = sys.argv[1]
if not filename.endswith(".npy"):
    print("Error : file is not a .npy file. Aborting")
    sys.exit()

data = np.load(filename)

# sum color channels for now
if len(data.shape) == 4:
    data = np.sum(data, axis=3)

factor = 1.0 / float(sys.argv[2]) if len(sys.argv) > 2 else 1
data = resize3d(data, factor)


if np.max(data) > 0:
    data = data / np.max(data)
else:
    print("Error : data has no positive value. Aborting.")
    sys.exit()

# default value of mean^0.5 is an heuristic
threshold = pow(np.mean(data), 0.5) if len(sys.argv) == 2 else sys.argv[2]
print("threshold : ", threshold)

data_shape = data.shape

output_file = open(filename.replace(".npy", ".obj"), "w")

i = 0
for z in range(data_shape[0]):
    print("converting depth layer ", z + 1, "/", data_shape[0])
    for x in range(data_shape[1]):
        for y in range(data_shape[2]):
            v = data[z, x, y]
            if v >= threshold:
                v = v / 2

                output_file.write(
                    "v "
                    + str(x)
                    + " "
                    + str(y)
                    + " "
                    + str(z - v)
                    + "\n"
                    + "v "
                    + str(x)
                    + " "
                    + str(y)
                    + " "
                    + str(z + v)
                    + "\n"
                    + "v "
                    + str(x)
                    + " "
                    + str(y - v)
                    + " "
                    + str(z)
                    + "\n"
                    + "v "
                    + str(x)
                    + " "
                    + str(y + v)
                    + " "
                    + str(z)
                    + "\n"
                    + "v "
                    + str(x - v)
                    + " "
                    + str(y)
                    + " "
                    + str(z)
                    + "\n"
                    + "v "
                    + str(x + v)
                    + " "
                    + str(y)
                    + " "
                    + str(z)
                    + "\n"
                    + "\n"
                    + "f "
                    + str(i + 1)
                    + " "
                    + str(i + 3)
                    + " "
                    + str(i + 5)
                    + "\n"
                    + "f "
                    + str(i + 1)
                    + " "
                    + str(i + 3)
                    + " "
                    + str(i + 6)
                    + "\n"
                    + "f "
                    + str(i + 1)
                    + " "
                    + str(i + 4)
                    + " "
                    + str(i + 5)
                    + "\n"
                    + "f "
                    + str(i + 1)
                    + " "
                    + str(i + 4)
                    + " "
                    + str(i + 6)
                    + "\n"
                    + "f "
                    + str(i + 2)
                    + " "
                    + str(i + 3)
                    + " "
                    + str(i + 5)
                    + "\n"
                    + "f "
                    + str(i + 2)
                    + " "
                    + str(i + 3)
                    + " "
                    + str(i + 6)
                    + "\n"
                    + "f "
                    + str(i + 2)
                    + " "
                    + str(i + 4)
                    + " "
                    + str(i + 5)
                    + "\n"
                    + "f "
                    + str(i + 2)
                    + " "
                    + str(i + 4)
                    + " "
                    + str(i + 6)
                    + "\n"
                    + "\n\n"
                )
                i += 6


output_file.close()



# #############################################################################
# simulation.py
# =================
# Authors :
# Yohann PERRON [yohann.perron@gmail.com]
# Eric BEZZAM [ebezzam@gmail.com]
# #############################################################################

from waveprop.simulation import FarFieldSimulator as FarFieldSimulator_wp
import torch


class FarFieldSimulator(FarFieldSimulator_wp):
    """
    LenslessPiCam-compatible wrapper for :py:class:`~waveprop.simulation.FarFieldSimulator` (source code on `GitHub <https://github.com/ebezzam/waveprop/blob/82dfb08b4db11c0c07ef00bdb59b5a769a49f0b3/waveprop/simulation.py#L11C11-L11C11>`__).
    """

    def __init__(
        self,
        object_height,
        scene2mask,
        mask2sensor,
        sensor,
        psf=None,
        output_dim=None,
        snr_db=None,
        max_val=255,
        device_conv="cpu",
        random_shift=False,
        is_torch=False,
        quantize=True,
        **kwargs
    ):
        """
        Parameters
        ----------
        psf : np.ndarray or torch.Tensor, optional.
            Point spread function. If not provided, return image at object plane.
        object_height : float or (float, float)
            Height of object in meters. Or range of values to randomly sample from.
        scene2mask : float
            Distance from scene to mask in meters.
        mask2sensor : float
            Distance from mask to sensor in meters.
        sensor : str
            Sensor name.
        snr_db : float, optional
            Signal-to-noise ratio in dB, by default None.
        max_val : int, optional
            Maximum value of image, by default 255.
        device_conv : str, optional
            Device to use for convolution (when using pytorch), by default "cpu".
        random_shift : bool, optional
            Whether to randomly shift the image, by default False.
        is_torch : bool, optional
            Whether to use pytorch, by default False.
        quantize : bool, optional
            Whether to quantize image, by default True.
        """

        if psf is not None:
            assert len(psf.shape) == 4, "PSF must be of shape (depth, height, width, channels)"

            if torch.is_tensor(psf):
                # drop depth dimension, and convert HWC to CHW
                psf = psf[0].movedim(-1, 0)
                assert psf.shape[0] == 1 or psf.shape[0] == 3, "PSF must have 1 or 3 channels"
            else:
                psf = psf[0]
                assert psf.shape[-1] == 1 or psf.shape[-1] == 3, "PSF must have 1 or 3 channels"

        super().__init__(
            object_height,
            scene2mask,
            mask2sensor,
            sensor,
            psf,
            output_dim,
            snr_db,
            max_val,
            device_conv,
            random_shift,
            is_torch,
            quantize,
            **kwargs
        )

        if psf is not None:
            if self.is_torch:
                assert (
                    self.psf.shape[0] == 1 or self.psf.shape[0] == 3
                ), "PSF must have 1 or 3 channels"
            else:
                assert (
                    self.psf.shape[-1] == 1 or self.psf.shape[-1] == 3
                ), "PSF must have 1 or 3 channels"

        # save all the parameters in a dict
        self.params = {
            "object_height": object_height,
            "scene2mask": scene2mask,
            "mask2sensor": mask2sensor,
            "sensor": sensor,
            "output_dim": output_dim,
            "snr_db": snr_db,
            "max_val": max_val,
            "device_conv": device_conv,
            "random_shift": random_shift,
            "is_torch": is_torch,
            "quantize": quantize,
        }
        self.params.update(kwargs)

    def get_psf(self):
        if self.is_torch:
            # convert CHW to HWC
            return self.psf.movedim(0, -1).unsqueeze(0)
        else:
            return self.psf[None, ...]

    # needs different name from parent class
    def set_point_spread_function(self, psf):
        """
        Set point spread function.

        Parameters
        ----------
        psf : np.ndarray or torch.Tensor
            Point spread function.
        """
        assert len(psf.shape) == 4, "PSF must be of shape (depth, height, width, channels)"

        if torch.is_tensor(psf):
            # convert HWC to CHW
            psf = psf[0].movedim(-1, 0)
            assert psf.shape[0] == 1 or psf.shape[0] == 3, "PSF must have 1 or 3 channels"
        else:
            psf = psf[0]
            assert psf.shape[-1] == 1 or psf.shape[-1] == 3, "PSF must have 1 or 3 channels"

        return super().set_psf(psf)

    def propagate_image(self, obj, return_object_plane=False):
        """
        Parameters
        ----------
        obj : np.ndarray or torch.Tensor
            Single image to propagate of format HWC.
        return_object_plane : bool, optional
            Whether to return object plane, by default False.
        """

        assert obj.shape[-1] == 1 or obj.shape[-1] == 3, "Image must have 1 or 3 channels"

        if self.is_torch:
            # channel in first dimension as expected by waveprop for pytorch
            obj = obj.moveaxis(-1, 0)
            res = super().propagate(obj, return_object_plane)
            if isinstance(res, tuple):
                res = res[0].moveaxis(-3, -1), res[1].moveaxis(-3, -1)
            else:
                res = res.moveaxis(-3, -1)
            return res
        else:
            # TODO: not tested, but normally don't need to move dimensions for numpy
            res = super().propagate(obj, return_object_plane)
            return res



import numpy as np
from lensless.hardware.utils import set_mask_sensor_distance
import hydra
import os
from datetime import datetime
from PIL import Image

SATURATION_THRESHOLD = 0.01


@hydra.main(version_base=None, config_path="../../configs", config_name="digicam_config")
def config_digicam(config):

    rpi_username = config.rpi.username
    rpi_hostname = config.rpi.hostname

    mask_sensor_distances = np.arange(9) * 0.1
    exposure_time = 5

    timestamp = datetime.now().strftime("%Y%m%d")

    for i in range(len(mask_sensor_distances)):

        print(f"Mask sensor distance: {mask_sensor_distances[i]}mm")
        mask_sensor_distance = mask_sensor_distances[i]

        # set the mask sensor distance
        set_mask_sensor_distance(mask_sensor_distance, rpi_username, rpi_hostname)

        good_exposure = False
        while not good_exposure:

            # measure PSF
            output_folder = f"adafruit_psf_{mask_sensor_distance}mm__{timestamp}"
            os.system(
                f"python scripts/remote_capture.py -cn capture_bayer output={output_folder} rpi.username={rpi_username} rpi.hostname={rpi_hostname} capture.exp={exposure_time}"
            )

            # check for saturation
            OUTPUT_FP = os.path.join(output_folder, "raw_data.png")
            # -- load picture to check for saturation
            img = np.array(Image.open(OUTPUT_FP))
            ratio = np.sum(img == 4095) / np.prod(img.shape)
            print(f"Saturation ratio: {ratio}")
            if ratio > SATURATION_THRESHOLD or ratio == 0:

                if ratio == 0:
                    print("Need to increase exposure time.")
                else:
                    print("Need to decrease exposure time.")

                # enter new exposure time from keyboard
                exposure_time = float(input("Enter new exposure time: "))

            else:
                good_exposure = True


if __name__ == "__main__":
    config_digicam()



# #############################################################################
# rfft_convolve.py
# =================
# Authors :
# Eric BEZZAM [ebezzam@gmail.com]
# Julien SAHLI [julien.sahli@epfl.ch]
# #############################################################################


"""
2D convolution in Fourier domain, with same real-valued kernel.
"""

import numpy as np
from scipy import fft
from scipy.fftpack import next_fast_len

try:
    import torch

    torch_available = True
except ImportError:
    torch_available = False


class RealFFTConvolve2D:
    def __init__(self, psf, dtype=None, pad=True, norm="ortho", rgb=None, **kwargs):
        """
        Linear operator that performs convolution in Fourier domain, and assumes
        real-valued signals.

        Parameters
        ----------
        psf :py:class:`~numpy.ndarray` or :py:class:`~torch.Tensor`
            Point spread function (PSF) that models forward propagation.
            Must be of shape (depth, height, width, channels) even if
            depth = 1 and channels = 1. You can use :py:func:`~lensless.io.load_psf`
            to load a PSF from a file such that it is in the correct format.
        dtype : float32 or float64
            Data type to use for optimization.
        pad : bool, optional
            Whether data needs to be padded prior to convolution. User may wish to
            optimize padded data and set this to False, as is done for :py:class:`~lensless.ADMM`.
            Defaults to True.
        norm : str, optional
            Normalization to use for FFT. Defaults to 'ortho'.
        """

        self.is_torch = False
        if torch_available and isinstance(psf, torch.Tensor):
            self.is_torch = True

        # prepare shapes for reconstruction

        assert (
            len(psf.shape) >= 4
        ), "Expected 4D PSF of shape ([batch], depth, width, height, channels)"
        self._use_3d = psf.shape[-4] != 1
        if rgb is None:
            self._is_rgb = psf.shape[-1] == 3
        else:
            self._is_rgb = rgb
        assert self._is_rgb or psf.shape[-1] == 1

        # save normalization
        self.norm = norm

        # set dtype
        if dtype is None:
            if self.is_torch:
                dtype = torch.float32
            else:
                dtype = np.float32
        self.dtype = dtype

        self.pad = pad  # Whether necessary to pad provided data
        self.set_psf(psf)

    def _crop(self, x):
        return x[
            ..., self._start_idx[0] : self._end_idx[0], self._start_idx[1] : self._end_idx[1], :
        ]

    def _pad(self, v):
        if len(v.shape) == 5:
            batch_size = v.shape[0]
            shape = [batch_size] + self._padded_shape
        elif len(v.shape) == 4:
            shape = self._padded_shape
        else:
            raise ValueError("Expected 4D or 5D tensor")

        if self.is_torch:
            vpad = torch.zeros(size=shape, dtype=v.dtype, device=v.device)
        else:
            vpad = np.zeros(shape).astype(v.dtype)
        vpad[
            ..., self._start_idx[0] : self._end_idx[0], self._start_idx[1] : self._end_idx[1], :
        ] = v
        return vpad

    def set_psf(self, psf):
        if self.is_torch:
            self._psf = psf.type(self.dtype)
        else:
            self._psf = psf.astype(self.dtype)

        self._psf_shape = np.array(self._psf.shape)

        # cropping / padding indexes
        self._padded_shape = 2 * self._psf_shape[-3:-1] - 1
        self._padded_shape = np.array([next_fast_len(i) for i in self._padded_shape])
        self._padded_shape = list(
            np.r_[self._psf_shape[-4], self._padded_shape, 3 if self._is_rgb else 1]
        )
        self._start_idx = (self._padded_shape[-3:-1] - self._psf_shape[-3:-1]) // 2
        self._end_idx = self._start_idx + self._psf_shape[-3:-1]

        # precompute filter in frequency domain
        if self.is_torch:
            self._H = torch.fft.rfft2(
                self._pad(self._psf), norm=self.norm, dim=(-3, -2), s=self._padded_shape[-3:-1]
            )
            self._Hadj = torch.conj(self._H)
            self._padded_data = (
                None  # This must be reinitialized each time to preserve differentiability
            )
        else:
            self._H = fft.rfft2(self._pad(self._psf), axes=(-3, -2), norm=self.norm)
            self._Hadj = np.conj(self._H)
            self._padded_data = np.zeros(self._padded_shape).astype(self.dtype)

    def convolve(self, x, return_fft=False):
        """
        Convolve with pre-computed FFT of provided PSF.
        """
        if self.pad:
            self._padded_data = self._pad(x)
        else:
            if self.is_torch:
                self._padded_data = x  # .type(self.dtype).to(self._psf.device)
            else:
                self._padded_data[:] = x  # .astype(self.dtype)

        if self.is_torch:
            conv_output = torch.fft.rfft2(self._padded_data, dim=(-3, -2)) * self._H

            if return_fft:
                return conv_output
            conv_output = torch.fft.ifftshift(
                torch.fft.irfft2(
                    conv_output,
                    dim=(-3, -2),
                    s=self._padded_shape[-3:-1],
                ),
                dim=(-3, -2),
            )

        else:
            conv_output = fft.rfft2(self._padded_data, axes=(-3, -2)) * self._H
            if return_fft:
                return conv_output
            conv_output = fft.ifftshift(
                fft.irfft2(
                    conv_output,
                    axes=(-3, -2),
                    s=self._padded_shape[-3:-1],
                ),
                axes=(-3, -2),
            )
        if self.pad:
            conv_output = self._crop(conv_output)

        # ensure shape stays the same
        assert conv_output.shape[-3:-1] == x.shape[-3:-1]
        return conv_output

    def deconvolve(self, y, return_fft=False):
        """
        Deconvolve with adjoint of pre-computed FFT of provided PSF.
        """
        if self.pad:
            self._padded_data = self._pad(y)
        else:
            if self.is_torch:
                self._padded_data = y  # .type(self.dtype).to(self._psf.device)
            else:
                self._padded_data[:] = y  # .astype(self.dtype)

        if self.is_torch:
            deconv_out = torch.fft.rfft2(self._padded_data, dim=(-3, -2)) * self._Hadj

            if return_fft:
                return deconv_out
            deconv_output = torch.fft.ifftshift(
                torch.fft.irfft2(
                    deconv_out,
                    dim=(-3, -2),
                    s=self._padded_shape[-3:-1],
                ),
                dim=(-3, -2),
            )

        else:
            deconv_output = fft.rfft2(self._padded_data, axes=(-3, -2)) * self._Hadj
            if return_fft:
                return deconv_output

            deconv_output = fft.ifftshift(
                fft.irfft2(
                    deconv_output,
                    axes=(-3, -2),
                    s=self._padded_shape[-3:-1],
                ),
                axes=(-3, -2),
            )

        if self.pad:
            deconv_output = self._crop(deconv_output)

        # ensure shape stays the same
        assert deconv_output.shape[-3:-1] == y.shape[-3:-1]
        return deconv_output



# #############################################################################
# train_learning_based.py
# =======================
# Authors :
# Yohann PERRON [yohann.perron@gmail.com]
# Eric BEZZAM [ebezzam@gmail.com]
# #############################################################################

"""
Base script for training learned reconstruction approaches.

Check out this page for configuration details:
https://github.com/LCAV/LenslessPiCam/tree/main/configs/train#training-physics-informed-reconstruction-models


"""

import wandb
import logging
import hydra
from hydra.utils import get_original_cwd, to_absolute_path
import os
import numpy as np
import time
from lensless.utils.image import shift_with_pad
from lensless.hardware.trainable_mask import prep_trainable_mask
from lensless import ADMM, UnrolledFISTA, UnrolledADMM, TrainableInversion, SVDeconvNet
from lensless.recon.multi_wiener import MultiWiener
from lensless.recon.integrated_background_sub import IntegratedBackgroundSub
from lensless.utils.dataset import (
    DiffuserCamMirflickr,
    DigiCamCelebA,
    HFDataset,
    MyDataParallel,
    simulate_dataset,
    HFSimulated,
)
from torch.utils.data import Subset
from lensless.recon.utils import create_process_network
from lensless.recon.utils import Trainer
import torch
from lensless.utils.io import save_image
from lensless.utils.plot import plot_image
import matplotlib.pyplot as plt
from lensless.recon.model_dict import load_model, download_model

# A logger for this file
log = logging.getLogger(__name__)


@hydra.main(version_base=None, config_path="../../configs/train", config_name="defaults")
def train_learned(config):

    if config.wandb_project is not None:
        # start a new wandb run to track this script
        wandb.init(
            # set the wandb project where this run will be logged
            project=config.wandb_project,
            # track hyperparameters and run metadata
            config=dict(config),
        )

    # set seed
    seed = config.seed
    torch.manual_seed(seed)
    np.random.seed(seed)
    generator = torch.Generator().manual_seed(seed)

    if config.start_delay is not None:
        # wait for this time before starting script
        delay = config.start_delay * 60
        start_time = time.time() + delay
        start_time = time.strftime("%H:%M:%S", time.localtime(start_time))
        print(f"\nScript will start at {start_time}")
        time.sleep(delay)

    save = config.save
    if save:
        save = os.getcwd()

    use_cuda = False

    # check torch_device is iterable
    if hasattr(config.torch_device, "__iter__") and not isinstance(config.torch_device, str):
        # assert every element is integer
        assert all(isinstance(x, int) for x in config.torch_device)
        device = f"cuda:{config.torch_device[0]}"
        device_ids = config.torch_device
    else:
        device = config.torch_device
        device_ids = None

    if "cuda" in device and torch.cuda.is_available():
        log.info(f"Using GPU for training. Main device : {device}")
        use_cuda = True
    else:
        log.info("Using CPU for training.")
        device = "cpu"
    if device_ids is not None:
        log.info(f"Using multiple GPUs : {device_ids}")

    # load dataset and create dataloader
    train_set = None
    test_set = None
    psf = None
    crop = None
    mask = None
    if "DiffuserCam" in config.files.dataset and config.files.huggingface_dataset is False:

        original_path = os.path.join(get_original_cwd(), config.files.dataset)
        psf_path = os.path.join(get_original_cwd(), config.files.psf)
        dataset = DiffuserCamMirflickr(
            dataset_dir=original_path,
            psf_path=psf_path,
            downsample=config.files.downsample,
            input_snr=config.files.input_snr,
        )
        dataset.psf = dataset.psf.to(device)
        # train-test split as in https://waller-lab.github.io/LenslessLearning/dataset.html
        # first 1000 files for test, the rest for training
        train_indices = dataset.allowed_idx[dataset.allowed_idx > 1000]
        test_indices = dataset.allowed_idx[dataset.allowed_idx <= 1000]
        if config.files.n_files is not None:
            train_indices = train_indices[: config.files.n_files]
            test_indices = test_indices[: config.files.n_files]

        train_set = Subset(dataset, train_indices)
        test_set = Subset(dataset, test_indices)

        # -- if learning mask
        mask = prep_trainable_mask(config, dataset.psf)

        psf = dataset.psf

    elif "celeba_adafruit" in config.files.dataset:

        dataset = DigiCamCelebA(
            data_dir=os.path.join(get_original_cwd(), config.files.dataset),
            celeba_root=config.files.celeba_root,
            psf_path=os.path.join(get_original_cwd(), config.files.psf),
            downsample=config.files.downsample,
            vertical_shift=config.files.vertical_shift,
            horizontal_shift=config.files.horizontal_shift,
            simulation_config=config.simulation,
            crop=config.files.crop,
            input_snr=config.files.input_snr,
        )
        crop = dataset.crop
        dataset.psf = dataset.psf.to(device)
        log.info(f"Data shape :  {dataset[0][0].shape}")

        if config.files.n_files is not None:
            dataset = Subset(dataset, np.arange(config.files.n_files))
            dataset.psf = dataset.dataset.psf

        # train-test split
        train_size = int((1 - config.files.test_size) * len(dataset))
        test_size = len(dataset) - train_size
        train_set, test_set = torch.utils.data.random_split(
            dataset, [train_size, test_size], generator=generator
        )

        # -- if learning mask
        downsample = config.files.downsample * 4  # measured files are 4x downsampled
        mask = prep_trainable_mask(config, dataset.psf, downsample=downsample)

        if mask is not None:
            # save original PSF
            psf_meas = dataset.psf.detach().cpu().numpy()[0, ...]
            plot_image(psf_meas, gamma=config.display.gamma)
            plt.savefig(os.path.join(save, "psf_meas_plot.png"))

            with torch.no_grad():
                psf = mask.get_psf().to(dataset.psf)

        else:

            psf = dataset.psf

    elif config.files.huggingface_dataset is True:

        split_train = "train"
        split_test = "test"
        if config.files.split_seed is not None:
            from datasets import load_dataset, concatenate_datasets

            seed = config.files.split_seed
            generator = torch.Generator().manual_seed(seed)

            # - combine train and test into single dataset
            train_split = "train"
            test_split = "test"
            if config.files.n_files is not None:
                train_split = f"train[:{config.files.n_files}]"
                test_split = f"test[:{config.files.n_files}]"
            train_dataset = load_dataset(
                config.files.dataset, split=train_split, cache_dir=config.files.cache_dir
            )
            test_dataset = load_dataset(
                config.files.dataset, split=test_split, cache_dir=config.files.cache_dir
            )
            dataset = concatenate_datasets([test_dataset, train_dataset])

            # - split into train and test
            train_size = int((1 - config.files.test_size) * len(dataset))
            test_size = len(dataset) - train_size
            split_train, split_test = torch.utils.data.random_split(
                dataset, [train_size, test_size], generator=generator
            )

        if config.files.hf_simulated:
            # simulate lensless by using measured PSF
            train_set = HFSimulated(
                huggingface_repo=config.files.dataset,
                split=split_train,
                n_files=config.files.n_files,
                psf=config.files.huggingface_psf,
                downsample=config.files.downsample,
                cache_dir=config.files.cache_dir,
                single_channel_psf=config.files.single_channel_psf,
                flipud=config.files.flipud,
                display_res=config.files.image_res,
                alignment=config.alignment,
                bg_snr_range=config.files.background_snr_range,  # TODO check if correct
                bg_fp=(
                    to_absolute_path(config.files.background_fp)
                    if config.files.background_fp is not None
                    else None
                ),
                input_snr=config.files.input_snr,
            )

        else:
            train_set = HFDataset(
                huggingface_repo=config.files.dataset,
                cache_dir=config.files.cache_dir,
                psf=config.files.huggingface_psf,
                single_channel_psf=config.files.single_channel_psf,
                split=split_train,
                display_res=config.files.image_res,
                rotate=config.files.rotate,
                flipud=config.files.flipud,
                flip_lensed=config.files.flip_lensed,
                downsample=config.files.downsample,
                downsample_lensed=config.files.downsample_lensed,
                alignment=config.alignment,
                save_psf=config.files.save_psf,
                n_files=config.files.n_files,
                simulation_config=config.simulation,
                force_rgb=config.files.force_rgb,
                simulate_lensless=config.files.simulate_lensless,
                random_flip=config.files.random_flip,
                bg_snr_range=config.files.background_snr_range,
                bg_fp=(
                    to_absolute_path(config.files.background_fp)
                    if config.files.background_fp is not None
                    else None
                ),
                input_snr=config.files.input_snr,
                psf_snr=config.files.psf_snr,
            )

        test_set = HFDataset(
            huggingface_repo=config.files.dataset,
            cache_dir=config.files.cache_dir,
            psf=config.files.huggingface_psf,
            single_channel_psf=config.files.single_channel_psf,
            split=split_test,
            display_res=config.files.image_res,
            rotate=config.files.rotate,
            flipud=config.files.flipud,
            flip_lensed=config.files.flip_lensed,
            downsample=config.files.downsample,
            downsample_lensed=config.files.downsample_lensed,
            alignment=config.alignment,
            save_psf=config.files.save_psf,
            n_files=config.files.n_files,
            simulation_config=config.simulation,
            bg_snr_range=config.files.background_snr_range,
            bg_fp=(
                to_absolute_path(config.files.background_fp)
                if config.files.background_fp is not None
                else None
            ),
            force_rgb=config.files.force_rgb,
            simulate_lensless=False,  # in general evaluate on measured (set to False)
            input_snr=config.files.input_snr,
        )

        if config.files.psf_snr is not None:
            # overwrite test set PSF with train set PSF
            test_set.psf = train_set.psf

        if train_set.multimask:
            # get first PSF for initialization
            if device_ids is not None:
                first_psf_key = list(train_set.psf.keys())[device_ids[0]]
            else:
                first_psf_key = list(train_set.psf.keys())[0]
            psf = train_set.psf[first_psf_key].to(device)
        else:
            psf = train_set.psf.to(device)
        crop = test_set.crop  # same for train set

        # -- if learning mask
        mask = prep_trainable_mask(config, psf)
        if mask is not None:
            assert not train_set.multimask

    else:

        train_set, test_set, mask = simulate_dataset(config, generator=generator)
        psf = train_set.psf
        crop = train_set.crop

    if not hasattr(train_set, "multimask"):
        train_set.multimask = False
    if not hasattr(test_set, "multimask"):
        test_set.multimask = False

    assert train_set is not None
    # if not hasattr(test_set, "psfs"):
    #     assert psf is not None

    # print info about PSF
    log.info(f"PSF shape : {psf.shape}")
    log.info(f"PSF min : {psf.min()}")
    log.info(f"PSF max : {psf.max()}")
    log.info(f"PSF dtype : {psf.dtype}")
    log.info(f"PSF norm : {psf.norm()}")

    if config.files.extra_eval is not None:
        # TODO only support Hugging Face DigiCam datasets for now
        extra_eval_sets = dict()
        for eval_set in config.files.extra_eval:

            extra_eval_sets[eval_set] = HFDataset(
                split="test",
                downsample=config.files.downsample,  # needs to be same size
                n_files=config.files.n_files,
                simulation_config=config.simulation,
                simulate_lensless=False,  # in general evaluate on measured
                **config.files.extra_eval[eval_set],
            )

    # reconstruct lensless with ADMM
    with torch.no_grad():
        if config.eval_disp_idx is not None:

            log.info("Reconstruction a few images with ADMM...")

            for i, _idx in enumerate(config.eval_disp_idx):

                flip_lr = None
                flip_ud = None
                return_items = test_set[_idx]
                lensless = return_items[0]
                lensed = return_items[1]
                if test_set.bg_sim is not None or test_set.measured_bg:
                    background = return_items[-1]
                if test_set.multimask or test_set.random_flip:
                    psf_recon = return_items[2]
                    psf_recon = psf_recon.to(device)
                else:
                    psf_recon = psf.clone()
                if test_set.random_flip:
                    flip_lr = return_items[3]
                    flip_ud = return_items[4]

                rotate_angle = False
                if config.files.random_rotate:
                    from lensless.utils.image import rotate_HWC

                    rotate_angle = np.random.uniform(
                        -config.files.random_rotate, config.files.random_rotate
                    )
                    print(f"Rotate angle : {rotate_angle}")
                    lensless = rotate_HWC(lensless, rotate_angle)
                    lensed = rotate_HWC(lensed, rotate_angle)
                    psf_recon = rotate_HWC(psf_recon, rotate_angle)

                shift = None
                if config.files.random_shifts:

                    shift = np.random.randint(
                        -config.files.random_shifts, config.files.random_shifts, 2
                    )
                    print(f"Shift : {shift}")
                    lensless = shift_with_pad(lensless, shift, axis=(1, 2))
                    lensed = shift_with_pad(lensed, shift, axis=(1, 2))
                    psf_recon = shift_with_pad(psf_recon, shift, axis=(1, 2))
                    shift = tuple(shift)

                if config.files.random_rotate or config.files.random_shifts:
                    save_image(psf_recon[0].cpu().numpy(), f"psf_{_idx}.png")

                save_image(lensed[0].cpu().numpy(), f"lensed_{_idx}.png")
                save_image(lensless[0].cpu().numpy(), f"lensless_raw_{_idx}.png")

                # Reconstruct and plot image
                reconstruct_save(
                    _idx,
                    config,
                    crop,
                    i,
                    lensed,
                    lensless,
                    psf_recon,
                    test_set,
                    "",
                    flip_lr,
                    flip_ud,
                    rotate_angle,
                    shift,
                )
                # save_image(lensed, f"lensed_{_idx}.png")
                if test_set.bg_sim is not None or test_set.measured_bg:
                    # Reconstruct and plot background subtracted image
                    reconstruct_save(
                        _idx,
                        config,
                        crop,
                        i,
                        lensed,
                        (lensless - background),
                        psf_recon,
                        test_set,
                        "subtraction_",
                        flip_lr,
                        flip_ud,
                        rotate_angle,
                        shift,
                    )
    log.info(f"Train test size : {len(train_set)}")
    log.info(f"Test test size : {len(test_set)}")

    start_time = time.time()

    # Load pre-process model
    pre_proc_delay = config.reconstruction.pre_process.delay
    pre_process, pre_process_name = create_process_network(
        config.reconstruction.pre_process.network,
        device=device,
        device_ids=device_ids,
        background_subtraction=config.reconstruction.integrated_background_unetres,
        input_background=config.reconstruction.unetres_input_background,
        # unet param
        depth=config.reconstruction.pre_process.depth,
        nc=config.reconstruction.pre_process.nc,
        # restormer param
        restormer_params=config.reconstruction.pre_process.restormer_params,
    )

    # Load post-process model
    post_process, post_process_name = create_process_network(
        config.reconstruction.post_process.network,
        device=device,
        device_ids=device_ids,
        concatenate_compensation=(
            config.reconstruction.compensation[-1]
            if config.reconstruction.compensation is not None
            else False
        ),
        # unet param
        depth=config.reconstruction.post_process.depth,
        nc=config.reconstruction.post_process.nc,
        # restormer param
        restormer_params=config.reconstruction.post_process.restormer_params,
    )
    post_proc_delay = config.reconstruction.post_process.delay

    if post_process is not None and config.reconstruction.post_process.train_last_layer:
        for name, param in post_process.named_parameters():
            if "m_tail" in name:
                param.requires_grad = True
            else:
                param.requires_grad = False
            # print(name, param.requires_grad, param.numel())

    # initialize pre- and post processor with another model
    if config.reconstruction.init_processors is not None:

        if "hf" in config.reconstruction.init_processors:
            param = config.reconstruction.init_processors.split(":")
            camera = param[1]
            dataset = param[2]
            model_name = param[3]
            model_path = download_model(camera=camera, dataset=dataset, model=model_name)

        elif "local" in config.reconstruction.init_processors:
            model_path = config.reconstruction.init_processors.split(":")[1]

        else:
            raise ValueError(f"{config.reconstruction.init_processors} is not a supported model")

        model_orig = load_model(
            model_path=model_path,
            psf=psf,
            device=device,
        )

        # -- replace pre-process
        if config.reconstruction.init_pre:
            params1 = model_orig.pre_process_model.named_parameters()
            params2 = pre_process.named_parameters()
            dict_params2 = dict(params2)
            for name1, param1 in params1:
                if name1 in dict_params2:
                    dict_params2[name1].data.copy_(param1.data)

        # -- replace post-process
        if config.reconstruction.init_post:
            params1_post = model_orig.post_process_model.named_parameters()
            params2_post = post_process.named_parameters()
            dict_params2_post = dict(params2_post)
            for name1, param1 in params1_post:
                if name1 in dict_params2_post:
                    dict_params2_post[name1].data.copy_(param1.data)

    # network for PSF
    psf_network = None
    if config.reconstruction.psf_network:
        # create UnetRes for PSF
        psf_network, _ = create_process_network(
            network="UnetRes",
            depth=len(config.reconstruction.psf_network),
            nc=config.reconstruction.psf_network,
            device=device,
            device_ids=device_ids,
        )

    # check/prepare background subtraction
    background_network = None
    if config.reconstruction.direct_background_subtraction:
        assert test_set.measured_bg and train_set.measured_bg
        assert config.reconstruction.learned_background_subtraction is False
    elif config.reconstruction.learned_background_subtraction:
        assert config.reconstruction.direct_background_subtraction is False
        assert test_set.measured_bg and train_set.measured_bg

        # create UnetRes for background subtraction
        background_network, _ = create_process_network(
            network="UnetRes",
            depth=len(config.reconstruction.learned_background_subtraction),
            nc=config.reconstruction.learned_background_subtraction,
            device=device,
            device_ids=device_ids,
        )
    elif config.reconstruction.integrated_background_subtraction:
        raise NotImplementedError("Integrated background subtraction not fully integrated yet")

        assert config.reconstruction.direct_background_subtraction is False
        assert config.reconstruction.learned_background_subtraction is False
        assert pre_process is None
        assert test_set.measured_bg and train_set.measured_bg

        background = train_set[0][-1]
        background_network = IntegratedBackgroundSub(
            in_channels=3,
            out_channels=3,
            input_shape=background.shape,
            nc=config.reconstruction.integrated_background_subtraction,
            down_subtraction=config.reconstruction.down_subtraction,
        )

    # create reconstruction algorithm
    if config.reconstruction.init is not None:
        assert config.reconstruction.init_processors is None

        param = config.reconstruction.init.split(":")
        assert len(param) == 4, "hf model requires following format: hf:camera:dataset:model_name"
        camera = param[1]
        dataset = param[2]
        model_name = param[3]
        model_path = download_model(camera=camera, dataset=dataset, model=model_name)
        recon = load_model(
            model_path,
            psf,
            device,
            device_ids=device_ids,
            train_last_layer=config.reconstruction.post_process.train_last_layer,
        )

    else:

        if config.reconstruction.method == "unrolled_fista":
            recon = UnrolledFISTA(
                psf,
                n_iter=config.reconstruction.unrolled_fista.n_iter,
                tk=config.reconstruction.unrolled_fista.tk,
                pad=True,
                learn_tk=config.reconstruction.unrolled_fista.learn_tk,
                pre_process=pre_process if pre_proc_delay is None else None,
                post_process=post_process if post_proc_delay is None else None,
                background_network=background_network,
                psf_network=psf_network,
                psf_residual=config.reconstruction.psf_residual,
                skip_unrolled=config.reconstruction.skip_unrolled,
                return_intermediate=(
                    True if config.unrolled_output_factor > 0 or config.pre_proc_aux > 0 else False
                ),
                compensation=config.reconstruction.compensation,
                compensation_residual=config.reconstruction.compensation_residual,
                direct_background_subtraction=config.reconstruction.direct_background_subtraction,
                integrated_background_subtraction=config.reconstruction.integrated_background_subtraction,
            )
        elif config.reconstruction.method == "unrolled_admm":
            recon = UnrolledADMM(
                psf,
                n_iter=config.reconstruction.unrolled_admm.n_iter,
                mu1=config.reconstruction.unrolled_admm.mu1,
                mu2=config.reconstruction.unrolled_admm.mu2,
                mu3=config.reconstruction.unrolled_admm.mu3,
                tau=config.reconstruction.unrolled_admm.tau,
                pre_process=pre_process if pre_proc_delay is None else None,
                post_process=post_process if post_proc_delay is None else None,
                background_network=background_network,
                psf_network=psf_network,
                psf_residual=config.reconstruction.psf_residual,
                skip_unrolled=config.reconstruction.skip_unrolled,
                return_intermediate=(
                    True if config.unrolled_output_factor > 0 or config.pre_proc_aux > 0 else False
                ),
                compensation=config.reconstruction.compensation,
                compensation_residual=config.reconstruction.compensation_residual,
                direct_background_subtraction=config.reconstruction.direct_background_subtraction,
                integrated_background_subtraction=config.reconstruction.integrated_background_subtraction,
            )
        elif config.reconstruction.method == "trainable_inv":
            assert config.trainable_mask.mask_type == "TrainablePSF"
            assert psf_network is None
            recon = TrainableInversion(
                psf,
                K=config.reconstruction.trainable_inv.K,
                pre_process=pre_process if pre_proc_delay is None else None,
                post_process=post_process if post_proc_delay is None else None,
                background_network=background_network,
                return_intermediate=(
                    True if config.unrolled_output_factor > 0 or config.pre_proc_aux > 0 else False
                ),
                direct_background_subtraction=config.reconstruction.direct_background_subtraction,
                integrated_background_subtraction=config.reconstruction.integrated_background_subtraction,
            )
        elif config.reconstruction.method == "svdeconvnet":
            assert config.trainable_mask.mask_type == "TrainablePSF"
            assert psf_network is None
            recon = SVDeconvNet(
                psf,
                K=config.reconstruction.svdeconvnet.K,
                pre_process=pre_process if pre_proc_delay is None else None,
                post_process=post_process if post_proc_delay is None else None,
                background_network=background_network,
                return_intermediate=(
                    True if config.unrolled_output_factor > 0 or config.pre_proc_aux > 0 else False
                ),
                direct_background_subtraction=config.reconstruction.direct_background_subtraction,
                integrated_background_subtraction=config.reconstruction.integrated_background_subtraction,
            )
        elif config.reconstruction.method == "multi_wiener":

            if config.files.single_channel_psf:
                psf = psf[..., 0].unsqueeze(-1)
                psf_channels = 1
            else:
                psf_channels = 3

            assert config.reconstruction.direct_background_subtraction is False, "Not supported"
            assert config.reconstruction.learned_background_subtraction is False, "Not supported"
            assert config.reconstruction.integrated_background_subtraction is False, "Not supported"
            assert psf_network is None, "Not supported"

            recon = MultiWiener(
                in_channels=3,
                out_channels=3,
                psf=psf,
                psf_channels=psf_channels,
                nc=config.reconstruction.multi_wiener.nc,
                pre_process=pre_process if pre_proc_delay is None else None,
            )

        else:
            raise ValueError(f"{config.reconstruction.method} is not a supported algorithm")

        if device_ids is not None:
            recon = MyDataParallel(recon, device_ids=device_ids)
        if use_cuda:
            recon.to(device)

    # constructing algorithm name by appending pre and post process
    algorithm_name = config.reconstruction.method
    if config.reconstruction.pre_process.network is not None:
        algorithm_name = pre_process_name + "_" + algorithm_name
    if config.reconstruction.post_process.network is not None:
        algorithm_name += "_" + post_process_name

    # print number of trainable parameters
    n_param = sum(p.numel() for p in recon.parameters() if p.requires_grad)
    if mask is not None:
        n_param += sum(p.numel() for p in mask.parameters() if p.requires_grad)
    log.info(f"Training model with {n_param} parameters")
    if pre_process is not None:
        n_param = sum(p.numel() for p in pre_process.parameters() if p.requires_grad)
        log.info(f"-- Pre-process model with {n_param} parameters")
    if post_process is not None:
        n_param = sum(p.numel() for p in post_process.parameters() if p.requires_grad)
        log.info(f"-- Post-process model with {n_param} parameters")
    if background_network is not None:
        n_param = sum(p.numel() for p in background_network.parameters() if p.requires_grad)
        log.info(f"-- Background subtraction model with {n_param} parameters")
    if psf_network is not None:
        n_param = sum(p.numel() for p in psf_network.parameters() if p.requires_grad)
        log.info(f"-- PSF network model with {n_param} parameters")
    if mask is not None:
        n_param = sum(p.numel() for p in mask.parameters() if p.requires_grad)
        log.info(f"-- Trainable mask with {n_param} parameters")

    log.info(f"Setup time : {time.time() - start_time} s")
    log.info(f"PSF shape : {psf.shape}")
    log.info(f"Results saved in {save}")
    trainer = Trainer(
        recon=recon,
        train_dataset=train_set,
        test_dataset=test_set,
        mask=mask,
        batch_size=config.training.batch_size,
        eval_batch_size=config.training.eval_batch_size,
        loss=config.loss,
        lpips=config.lpips,
        l1_mask=config.trainable_mask.L1_strength,
        optimizer=config.optimizer,
        skip_NAN=config.training.skip_NAN,
        algorithm_name=algorithm_name,
        metric_for_best_model=config.training.metric_for_best_model,
        save_every=config.training.save_every,
        gamma=config.display.gamma,
        logger=log,
        crop=crop if config.training.crop_preloss else None,
        pre_process=pre_process,
        pre_process_delay=pre_proc_delay,
        pre_process_freeze=config.reconstruction.pre_process.freeze,
        pre_process_unfreeze=config.reconstruction.pre_process.unfreeze,
        post_process=post_process,
        post_process_delay=post_proc_delay,
        post_process_freeze=config.reconstruction.post_process.freeze,
        post_process_unfreeze=config.reconstruction.post_process.unfreeze,
        clip_grad=config.training.clip_grad,
        unrolled_output_factor=config.unrolled_output_factor,
        pre_proc_aux=config.pre_proc_aux,
        extra_eval_sets=extra_eval_sets if config.files.extra_eval is not None else None,
        use_wandb=True if config.wandb_project is not None else False,
        n_epoch=config.training.epoch,
        random_rotate=config.files.random_rotate,
        random_shift=config.files.random_shifts,
    )

    trainer.train(n_epoch=config.training.epoch, save_pt=save, disp=config.eval_disp_idx)

    log.info(f"Results saved in {save}")


def reconstruct_save(
    _idx,
    config,
    crop,
    i,
    lensed,
    lensless,
    psf_recon,
    test_set,
    fp,
    flip_lr,
    flip_ud,
    rotate_angle,
    shift,
):
    recon = ADMM(psf_recon)

    recon.set_data(lensless.to(psf_recon.device))
    # recon.set_data(torch.from_numpy(lensless).to(psf_recon.device))
    res = recon.apply(disp_iter=None, plot=False, n_iter=10)
    res_np = res[0].cpu().numpy()
    res_np = res_np / res_np.max()
    lensed_np = lensed[0]  # .cpu().numpy()

    # -- plot lensed and res on top of each other
    cropped = False
    if hasattr(test_set, "alignment"):
        if test_set.alignment is not None:
            res_np = test_set.extract_roi(
                res_np,
                axis=(0, 1),
                flip_lr=flip_lr,
                flip_ud=flip_ud,
                rotate_aug=rotate_angle,
                shift_aug=shift,
            )
        else:
            res_np, lensed_np = test_set.extract_roi(
                res_np,
                lensed=lensed_np,
                axis=(0, 1),
                flip_lr=flip_lr,
                flip_ud=flip_ud,
                rotate_aug=rotate_angle,
                shift_aug=shift,
            )
        cropped = True

    elif config.training.crop_preloss:
        assert crop is not None
        assert flip_lr is None and flip_ud is None

        res_np = res_np[
            crop["vertical"][0] : crop["vertical"][1],
            crop["horizontal"][0] : crop["horizontal"][1],
        ]
        lensed_np = lensed_np[
            crop["vertical"][0] : crop["vertical"][1],
            crop["horizontal"][0] : crop["horizontal"][1],
        ]
        cropped = True

    if cropped and i == 0:
        log.info(f"Cropped shape :  {res_np.shape}")

    save_image(res_np, f"lensless_recon_{fp}{_idx}.png")

    plt.figure()
    plt.imshow(lensed_np, alpha=0.4)
    plt.imshow(res_np, alpha=0.7)
    plt.savefig(f"overlay_lensed_recon_{fp}{_idx}.png")


if __name__ == "__main__":
    train_learned()



from waveprop.dataset_util import SimulatedDatasetFolder
import hydra
from hydra.utils import to_absolute_path
import os
from lensless.utils.io import load_psf
import torch
from torchvision.transforms import ToTensor
import time


@hydra.main(
    version_base=None,
    config_path="../../configs/simulate",
    config_name="simulate_torch_custom_dataset",
)
def simulate(config):

    dataset = to_absolute_path(config.files.dataset)
    if not os.path.isdir(dataset):
        print(f"No dataset found at {dataset}")
        try:
            from torchvision.datasets.utils import download_and_extract_archive, download_url
        except ImportError:
            exit()
        msg = "Do you want to download the sample CelebA dataset (764KB)?"

        # default to yes if no input is given
        valid = input("%s (Y/n) " % msg).lower() != "n"
        if valid:
            url = "https://drive.switch.ch/index.php/s/Q5OdDQMwhucIlt8/download"
            filename = "celeb_mini.zip"
            download_and_extract_archive(
                url, os.path.dirname(dataset), filename=filename, remove_finished=True
            )

    psf_fp = to_absolute_path(config.files.psf)
    assert os.path.exists(psf_fp), f"PSF {psf_fp} does not exist."

    image_ext = config.files.image_ext
    n_files = config.files.n_files
    batch_size = config.files.batch_size
    downsample = config.simulation.downsample
    device_conv = config.device_conv

    # check if gpu is available
    if device_conv == "cuda" and torch.cuda.is_available():
        print("Using GPU for convolution.")
        device_conv = "cuda"
    else:
        print("Using CPU for convolution.")
        device_conv = "cpu"

    # load PSF
    psf = load_psf(psf_fp, downsample=downsample)
    psf_tensor = ToTensor()(psf[0])  # first depth

    # create Pytorch dataset and dataloader
    ds = SimulatedDatasetFolder(
        path=dataset,
        iamge_ext=image_ext,
        n_files=n_files,
        psf=psf_tensor,
        device_conv=device_conv,
        **config.simulation,
    )
    ds_loader = torch.utils.data.DataLoader(dataset=ds, batch_size=batch_size, shuffle=True)

    # loop over batches
    start_time = time.time()
    for i, (x, target) in enumerate(ds_loader):

        if i == 0:
            print("Batch shape : ", x.shape)
            print("Target shape : ", target.shape)
            print("Batch device : ", x.device)

    print("Time per batch : ", (time.time() - start_time) / len(ds_loader))
    print(f"Went through {len(ds_loader)} batches.")


if __name__ == "__main__":
    simulate()



import hydra
from lensless.hardware.utils import set_mask_sensor_distance


@hydra.main(version_base=None, config_path="../../configs", config_name="digicam_config")
def config_digicam(config):

    rpi_username = config.rpi.username
    rpi_hostname = config.rpi.hostname

    # set mask to sensor distance
    set_mask_sensor_distance(config.z, rpi_username, rpi_hostname, max_distance=40)


if __name__ == "__main__":
    config_digicam()



import os
import hydra
from hydra.utils import to_absolute_path
import subprocess
import numpy as np
import time
from pprint import pprint
from lensless.utils.plot import plot_image, pixel_histogram
from lensless.utils.io import save_image
from lensless.utils.image import resize, print_image_info
import cv2
import matplotlib.pyplot as plt
from lensless import FISTA, ADMM
from lensless.hardware.utils import check_username_hostname, display
from lensless.utils.io import load_image, load_psf
import omegaconf
from lensless.hardware.slm import set_programmable_mask, adafruit_sub2full
from lensless.hardware.trainable_mask import AdafruitLCD
from torch import from_numpy


@hydra.main(version_base=None, config_path="../configs", config_name="demo")
def demo(config):

    check_username_hostname(config.rpi.username, config.rpi.hostname)
    RPI_USERNAME = config.rpi.username
    RPI_HOSTNAME = config.rpi.hostname

    display_fp = to_absolute_path(config.fp)
    if config.save:
        if config.output is not None:
            # make sure output directory exists
            os.makedirs(config.output, exist_ok=True)
            save = config.output
        else:
            save = os.getcwd()
    else:
        save = False

    # 1) Copy file to Raspberry Pi
    print("\nCopying over picture...")
    display(fp=display_fp, rpi_username=RPI_USERNAME, rpi_hostname=RPI_HOSTNAME, **config.display)

    # 2) (If DigiCam) set mask pattern
    mask = None
    flipud = False
    if isinstance(config.camera.psf, omegaconf.dictconfig.DictConfig):
        print("\nSetting mask pattern...")
        np.random.seed(config.camera.psf.seed % (2**32 - 1))
        mask_vals = np.random.uniform(0, 1, config.camera.psf.mask_shape)
        full_pattern = adafruit_sub2full(
            mask_vals,
            center=config.camera.psf.mask_center,
        )
        set_programmable_mask(
            full_pattern,
            device=config.camera.psf.device,
            rpi_username=RPI_USERNAME,
            rpi_hostname=RPI_HOSTNAME,
        )
        mask_vals_torch = from_numpy(mask_vals.astype(np.float32))
        flipud = config.camera.psf.flipud
        mask = AdafruitLCD(
            initial_vals=mask_vals_torch,
            sensor=config.capture.sensor,
            slm=config.camera.psf.device,
            downsample=config.recon.downsample,
            flipud=flipud,
        )

    # 3) Take picture
    time.sleep(config.capture.delay)  # for picture to display

    remote_fn = "remote_capture"
    print("\nTaking picture...")
    pic_command = (
        f"{config.rpi.python} {config.capture.script} sensor={config.capture.sensor} bayer={config.capture.bayer} fn={remote_fn} exp={config.capture.exp} iso={config.capture.iso} "
        f"config_pause={config.capture.config_pause} sensor_mode={config.capture.sensor_mode} nbits_out={config.capture.nbits_out} "
        f"legacy={config.capture.legacy} rgb={config.capture.rgb} gray={config.capture.gray} "
    )
    if config.capture.nbits > 8:
        pic_command += " sixteen=True"
    if config.capture.down:
        pic_command += f" down={config.capture.down}"
    if config.capture.awb_gains:
        pic_command += f" awb_gains=[{config.capture.awb_gains[0]},{config.capture.awb_gains[1]}]"

    print(f"COMMAND : {pic_command}")
    ssh = subprocess.Popen(
        ["ssh", "%s@%s" % (RPI_USERNAME, RPI_HOSTNAME), pic_command],
        shell=False,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    result = ssh.stdout.readlines()
    error = ssh.stderr.readlines()

    if (
        error != [] and config.capture.legacy
    ):  # new camera software seems to return error even if it works
        print("ERROR: %s" % error)
        return
    if result == []:
        error = ssh.stderr.readlines()
        print("ERROR: %s" % error)
        return
    else:
        result = [res.decode("UTF-8") for res in result]
        result = [res for res in result if len(res) > 3]
        result_dict = dict()
        for res in result:
            _key = res.split(":")[0].strip()
            _val = "".join(res.split(":")[1:]).strip()
            result_dict[_key] = _val
        # result_dict = dict(map(lambda s: map(str.strip, s.split(":")), result))
        print("COMMAND OUTPUT : ")
        pprint(result_dict)

    if (
        "RPi distribution" in result_dict.keys()
        and "bullseye" in result_dict["RPi distribution"]
        and not config.capture.legacy
    ):
        assert (
            not config.capture.rgb or not config.capture.gray
        ), "RGB and gray not supported for RPi HQ sensor"

        if config.capture.bayer:

            assert config.capture.down is None

            # copy over DNG file
            remotefile = f"~/{remote_fn}.dng"
            localfile = os.path.join(save, f"{config.capture.raw_data_fn}.dng")
            print(f"\nCopying over picture as {localfile}...")
            os.system('scp "%s@%s:%s" %s' % (RPI_USERNAME, RPI_HOSTNAME, remotefile, localfile))

            img = load_image(
                localfile,
                verbose=True,
                bayer=config.capture.bayer,
                nbits_out=config.capture.nbits_out,
            )

            # print image properties
            print_image_info(img)

            # save as PNG
            png_out = f"{config.capture.raw_data_fn}.png"
            print(f"Saving RGB file as: {png_out}")
            cv2.imwrite(png_out, cv2.cvtColor(img, cv2.COLOR_RGB2BGR))

        else:

            remotefile = f"~/{remote_fn}.png"
            localfile = f"{config.capture.raw_data_fn}.png"
            if save:
                localfile = os.path.join(save, localfile)
            print(f"\nCopying over picture as {localfile}...")
            os.system('scp "%s@%s:%s" %s' % (RPI_USERNAME, RPI_HOSTNAME, remotefile, localfile))

            img = load_image(localfile, verbose=True)

    # legacy software running on RPi
    else:
        # copy over file
        # more pythonic? https://stackoverflow.com/questions/250283/how-to-scp-in-python
        remotefile = f"~/{remote_fn}.png"
        localfile = f"{config.capture.raw_data_fn}.png"
        if save:
            localfile = os.path.join(save, localfile)
        print(f"\nCopying over picture as {localfile}...")
        os.system('scp "%s@%s:%s" %s' % (RPI_USERNAME, RPI_HOSTNAME, remotefile, localfile))

        if config.capture.rgb or config.capture.gray:
            img = load_image(localfile, verbose=True)

        else:

            if config.capture.bayer:
                red_gain = config.camera.red_gain
                blue_gain = config.camera.blue_gain
            else:
                red_gain = None
                blue_gain = None

            # load image
            print("\nLoading picture...")

            img = load_image(
                localfile,
                verbose=True,
                bayer=config.capture.bayer,
                blue_gain=blue_gain,
                red_gain=red_gain,
                nbits_out=config.capture.nbits_out,
            )

            # write RGB data
            if not config.capture.bayer:
                cv2.imwrite(localfile, cv2.cvtColor(img, cv2.COLOR_RGB2BGR))

    # plot image and histogram (useful to check clipping)
    ax = plot_image(img, gamma=config.capture.gamma)
    ax.set_title("Raw data")
    if save:
        plt.savefig(os.path.join(save, "raw.png"))
    pixel_histogram(img)
    if save:
        plt.savefig(os.path.join(save, "histogram.png"))

    # 4) Reconstruct

    # -- prepare data
    if mask is not None:
        psf = mask.get_psf().detach().numpy()
        bg = np.zeros(psf.shape[-1])
    else:
        psf, bg = load_psf(
            to_absolute_path(config.camera.psf),
            downsample=config.recon.downsample,
            return_float=True,
            return_bg=True,
            dtype=config.recon.dtype,
        )
        psf = np.array(psf, dtype=config.recon.dtype)
    ax = plot_image(psf[0], gamma=config.recon.gamma)
    ax.set_title("PSF")
    if save:
        plt.savefig(os.path.join(save, "psf.png"))

    data = np.array(img, dtype=config.recon.dtype)
    data -= bg
    data = np.clip(data, a_min=0, a_max=data.max())

    if len(data.shape) == 3:
        data = data[np.newaxis, :, :, :]
    elif len(data.shape) == 2:
        data = data[np.newaxis, :, :, np.newaxis]

    if data.shape != psf.shape:
        # in DiffuserCam dataset, images are already reshaped
        data = resize(data, shape=psf.shape)
    data /= np.linalg.norm(data.ravel())
    data = np.array(data, dtype=config.recon.dtype)

    if config.recon.use_torch:
        import torch

        if config.recon.dtype == "float32":
            torch_dtype = torch.float32
        elif config.recon.dtype == "float64":
            torch_dtype = torch.float64
        else:
            raise ValueError("dtype must be float32 or float64")

        psf = torch.from_numpy(psf).type(torch_dtype).to(config.recon.torch_device)
        data = torch.from_numpy(data).type(torch_dtype).to(config.recon.torch_device)
        if flipud:
            data = torch.rot90(data, dims=(-3, -2), k=2)
    else:
        if flipud:
            data = np.rot90(data, k=2, axes=(-3, -2))

    # -- apply algo
    start_time = time.time()

    if config.recon.algo == "fista":
        algo_params = config.recon.fista
        recon = FISTA(
            psf,
            **algo_params,
        )
    elif config.recon.algo == "admm":
        algo_params = config.recon.admm
        recon = ADMM(
            psf,
            **algo_params,
        )
    else:
        raise ValueError(f"Unsupported algorithm: {config.recon.algo}")

    recon.set_data(data)
    res = recon.apply(
        gamma=config.recon.gamma,
        save=save,
        plot=config.plot,
        disp_iter=algo_params["disp_iter"],
    )
    print(f"Processing time : {time.time() - start_time} s")

    if config.plot:
        final_image = res[0]
    else:
        final_image = res

    # save final image ax
    if save:

        # take first depth
        final_image = final_image[0]
        if config.recon.use_torch:
            img = final_image.cpu().numpy()
        else:
            img = final_image

        if config.postproc.crop_hor is not None:
            img = img[
                :,
                int(config.postproc.crop_hor[0] * img.shape[1]) : int(
                    config.postproc.crop_hor[1] * img.shape[1]
                ),
            ]
        if config.postproc.crop_vert is not None:
            img = img[
                int(config.postproc.crop_vert[0] * img.shape[0]) : int(
                    config.postproc.crop_vert[1] * img.shape[0]
                ),
                :,
            ]

        output_fp = os.path.join(save, "reconstructed.png")
        save_image(img, output_fp)

    # clean up
    os.remove(localfile)

    if config.plot:
        plt.show()

    return save


if __name__ == "__main__":
    demo()



# #############################################################################
# trainable_mask.py
# ==================
# Authors :
# Yohann PERRON [yohann.perron@gmail.com]
# Eric BEZZAM [ebezzam@gmail.com]
# #############################################################################

import abc
import omegaconf
import numpy as np
import torch
from lensless.utils.image import is_grayscale, rgb2gray
from lensless.hardware.slm import full2subpattern, get_programmable_mask, get_intensity_psf
from lensless.hardware.sensor import VirtualSensor
from waveprop.devices import slm_dict
from lensless.hardware.mask import CodedAperture


class TrainableMask(torch.nn.Module, metaclass=abc.ABCMeta):
    """
    Abstract class for defining trainable masks.

    The following abstract methods need to be defined:

    - :py:class:`~lensless.hardware.trainable_mask.TrainableMask.get_psf`: returning the PSF of the mask.
    - :py:class:`~lensless.hardware.trainable_mask.TrainableMask.project`: projecting the mask parameters to a valid space (should be a subspace of [0,1]).

    """

    def __init__(self, optimizer="Adam", lr=1e-3, **kwargs):
        """
        Base constructor. Derived constructor may define new state variables

        Parameters
        ----------
        optimizer : str, optional
            Optimizer to use for updating the mask parameters, by default "Adam"
        lr : float, optional
            Learning rate for the mask parameters, by default 1e-3
        """
        super().__init__()
        self._optimizer = optimizer
        self._lr = lr
        self._counter = 0

    def _set_optimizer(self, param):
        """Set the optimizer for the mask parameters."""
        self._optimizer = getattr(torch.optim, self._optimizer)(param, lr=self._lr)

    @abc.abstractmethod
    def get_psf(self):
        """
        Abstract method for getting the PSF of the mask. Should be fully compatible with pytorch autograd.

        Returns
        -------
        :py:class:`~torch.Tensor`
            The PSF of the mask.
        """
        raise NotImplementedError

    def update_mask(self):
        """Update the mask parameters. According to externaly updated gradiants."""
        self._optimizer.step()
        self._optimizer.zero_grad(set_to_none=True)
        self.project()
        self._counter += 1

    @abc.abstractmethod
    def project(self):
        """Abstract method for projecting the mask parameters to a valid space (should be a subspace of [0,1])."""
        raise NotImplementedError


class TrainablePSF(TrainableMask):
    # class TrainablePSF(torch.nn.Module, TrainableMask):
    """
    Class for defining an object that directly optimizes the PSF, without any constraints on what can be realized physically.

    Parameters
    ----------
    grayscale : bool, optional
        Whether mask should be returned as grayscale when calling :py:class:`~lensless.hardware.trainable_mask.TrainableMask.get_psf`.
        Otherwise PSF will be returned as RGB. By default False.
    """

    def __init__(self, initial_psf, grayscale=False, **kwargs):
        super().__init__(**kwargs)
        self._psf = torch.nn.Parameter(initial_psf)
        initial_param = [self._psf]
        self._set_optimizer(initial_param)

        # checks
        assert len(initial_psf.shape) == 4, "Mask must be of shape (depth, height, width, channels)"
        self.grayscale = grayscale
        self._is_grayscale = is_grayscale(initial_psf)
        if grayscale:
            assert self._is_grayscale, "PSF must be grayscale"

    def get_psf(self):
        if self._is_grayscale:
            if self.grayscale:
                # simulation in grayscale
                return self._psf
            else:
                # replicate to 3 channels
                return self._psf.expand(-1, -1, -1, 3)
        else:
            # assume RGB
            return self._psf

    def project(self):
        self._psf.data = torch.clamp(self._psf, 0, 1)


class AdafruitLCD(TrainableMask):
    def __init__(
        self,
        initial_vals,
        sensor,
        slm,
        train_mask_vals=True,
        color_filter=None,
        rotate=None,
        flipud=False,
        use_waveprop=False,
        vertical_shift=None,
        horizontal_shift=None,
        scene2mask=None,
        mask2sensor=None,
        deadspace=True,
        downsample=None,
        min_val=0,
        **kwargs,
    ):
        """
        Parameters
        ----------
        initial_vals : :py:class:`~torch.Tensor`
            Initial mask parameters.
        sensor : :py:class:`~lensless.hardware.sensor.VirtualSensor`
            Sensor object.
        slm_param : :py:class:`~lensless.hardware.slm.SLMParam`
            SLM parameters.
        rotate : float, optional
            Rotation angle in degrees, by default None.
        flipud : bool, optional
            Whether to flip the mask vertically, by default False.
        use_waveprop : bool, optional
            Whether to use wave propagation for simulating PSF. If False, PSF will simply be intensity of mask pattern, by default False.
        vertical_shift : int, optional
            Vertical shift of the mask, by default None.
        horizontal_shift : int, optional
            Horizontal shift of the mask, by default None.
        scene2mask : :py:class:`~torch.Tensor`, optional
            Distance from scene to mask. Used for wave propagation, by default None.
        mask2sensor : :py:class:`~torch.Tensor`, optional
            Distance from mask to sensor. Used for wave propagation, by default None.
        downsample : int, optional
            Downsample factor, by default None.
        min_val : float, optional
            Minimum value for mask weights, by default 0.
        """

        super().__init__(**kwargs)

        self.train_mask_vals = train_mask_vals
        if train_mask_vals:
            self._vals = torch.nn.Parameter(initial_vals)
        else:
            self._vals = initial_vals

        initial_param = None
        if color_filter is not None:
            self._color_filter = torch.nn.Parameter(color_filter)
            if train_mask_vals:
                initial_param = [self._vals, self._color_filter]
            else:
                initial_param = [self._color_filter]
        else:
            assert (
                train_mask_vals
            ), "If color filter is not trainable, mask values must be trainable"
            initial_param = [self._vals]
            self._color_filter = None
        assert initial_param is not None, "Initial parameters must be set"

        # set optimizer
        self._set_optimizer(initial_param)

        self.slm_param = slm_dict[slm]
        self.device = slm
        self.sensor = VirtualSensor.from_name(sensor, downsample=downsample)
        self.rotate = rotate
        self.flipud = flipud
        self.use_waveprop = use_waveprop
        self.scene2mask = scene2mask
        self.mask2sensor = mask2sensor
        self.deadspace = deadspace
        self.vertical_shift = vertical_shift
        self.horizontal_shift = horizontal_shift
        self.min_val = min_val
        if downsample is not None and vertical_shift is not None:
            self.vertical_shift = vertical_shift // downsample
        if downsample is not None and horizontal_shift is not None:
            self.horizontal_shift = horizontal_shift // downsample
        if self.use_waveprop:
            assert self.scene2mask is not None
            assert self.mask2sensor is not None

    def get_psf(self):

        mask = get_programmable_mask(
            vals=self._vals,
            sensor=self.sensor,
            slm_param=self.slm_param,
            rotate=self.rotate,
            flipud=self.flipud,
            color_filter=self._color_filter,
            deadspace=self.deadspace,
        )

        if self.vertical_shift is not None:
            mask = torch.roll(mask, self.vertical_shift, dims=1)

        if self.horizontal_shift is not None:
            mask = torch.roll(mask, self.horizontal_shift, dims=2)

        psf_in = get_intensity_psf(
            mask=mask,
            sensor=self.sensor,
            waveprop=self.use_waveprop,
            scene2mask=self.scene2mask,
            mask2sensor=self.mask2sensor,
        )

        # add first dimension (depth)
        psf_in = psf_in.unsqueeze(0)

        # move channels to last dimension
        psf_in = psf_in.permute(0, 2, 3, 1)

        # flip mask
        psf_in = torch.flip(psf_in, dims=[-3, -2])

        # normalize
        psf_in = psf_in / psf_in.norm()

        return psf_in

    def project(self):
        if self.train_mask_vals:
            self._vals.data = torch.clamp(self._vals, self.min_val, 1)
        if self._color_filter is not None:
            self._color_filter.data = torch.clamp(self._color_filter, 0, 1)
            # normalize each row to 1
            self._color_filter.data = self._color_filter / self._color_filter.sum(
                dim=[1, 2]
            ).unsqueeze(-1).unsqueeze(-1)


class TrainableCodedAperture(TrainableMask):
    def __init__(
        self,
        sensor_name,
        downsample=None,
        binary=True,
        torch_device="cuda",
        **kwargs,
    ):
        """
        TODO: Distinguish between separable and non-separable.
        """

        # 1) call base constructor so parameters can be set
        super().__init__(**kwargs)

        # 2) initialize mask
        assert "distance_sensor" in kwargs, "Distance to sensor must be specified"
        assert "method" in kwargs, "Method must be specified."
        assert "n_bits" in kwargs, "Number of bits must be specified."
        self._mask_obj = CodedAperture.from_sensor(
            sensor_name,
            downsample,
            is_torch=True,
            torch_device=torch_device,
            **kwargs,
        )

        # 3) set learnable parameters (should be immediate attributes of the class)
        self._row = None
        self._col = None
        self._mask = None
        if self._mask_obj.row is not None:
            # separable
            self.separable = True
            self._row = torch.nn.Parameter(self._mask_obj.row)
            self._col = torch.nn.Parameter(self._mask_obj.col)
            initial_param = [self._row, self._col]
        else:
            # non-separable
            self.separable = False
            self._mask = torch.nn.Parameter(self._mask_obj.mask)
            initial_param = [self._mask]
        self.binary = binary

        # 4) set optimizer
        self._set_optimizer(initial_param)

        # 5) compute PSF
        self._psf = None
        self.project()

    def get_psf(self):
        return self._psf

    def project(self):
        with torch.no_grad():
            if self.separable:
                self._row.data = torch.clamp(self._row, 0, 1)
                self._col.data = torch.clamp(self._col, 0, 1)
                if self.binary:
                    self._row.data = torch.round(self._row)
                    self._col.data = torch.round(self._col)
            else:
                self._mask.data = torch.clamp(self._mask, 0, 1)
                if self.binary:
                    self._mask.data = torch.round(self._mask)

        # recompute PSF
        self._mask_obj.create_mask(self._row, self._col, mask=self._mask)
        self._mask_obj.compute_psf()
        self._psf = self._mask_obj.psf.unsqueeze(0)
        self._psf = self._psf / self._psf.norm()


"""
Utility functions to help prepare trainable masks.
"""

mask_type_to_class = {
    "AdafruitLCD": AdafruitLCD,
    "TrainablePSF": TrainablePSF,
    "TrainableCodedAperture": TrainableCodedAperture,
    "TrainableHeightVarying": None,
    "TrainableMultiLensArray": None,
}


def prep_trainable_mask(config, psf=None, downsample=None):

    mask = None
    color_filter = None
    downsample = config["files"]["downsample"] if downsample is None else downsample
    mask_type = config["trainable_mask"]["mask_type"]
    if mask_type is not None:
        assert mask_type in mask_type_to_class.keys(), (
            f"Trainable mask type {mask_type} not supported. "
            f"Supported types are {mask_type_to_class.keys()}"
        )
        mask_class = mask_type_to_class[mask_type]

        # -- trainable mask object
        if isinstance(config["trainable_mask"]["initial_value"], omegaconf.dictconfig.DictConfig):

            # from mask config
            mask = mask_class(
                # mask = TrainableCodedAperture(
                sensor_name=config.simulation.sensor,
                downsample=downsample,
                distance_sensor=config.simulation.mask2sensor,
                optimizer=config.trainable_mask.optimizer,
                lr=config.trainable_mask.mask_lr,
                binary=config.trainable_mask.binary,
                torch_device=config.torch_device,
                **config.trainable_mask.initial_value,
            )

        else:

            if config["trainable_mask"]["initial_value"] == "random":
                if psf is not None:
                    initial_mask = torch.rand_like(psf)
                else:
                    sensor = VirtualSensor.from_name(
                        config["simulation"]["sensor"], downsample=downsample
                    )
                    resolution = sensor.resolution
                    initial_mask = torch.rand((1, *resolution, 3))

            elif config["trainable_mask"]["initial_value"] == "psf":

                if config["reconstruction"]["method"] == "svdeconvnet":
                    # learnable PSF as part of SVDeconvNet
                    n_psf = config["reconstruction"]["svdeconvnet"]["K"] ** 2
                    initial_mask = psf.repeat(n_psf, 1, 1, 1)
                else:
                    initial_mask = psf.clone()

            # if file ending with "npy"
            elif config["trainable_mask"]["initial_value"].endswith("npy"):

                pattern = np.load(config["trainable_mask"]["initial_value"])

                initial_mask = full2subpattern(
                    pattern=pattern,
                    shape=config["trainable_mask"]["ap_shape"],
                    center=config["trainable_mask"]["ap_center"],
                    slm=config["trainable_mask"]["slm"],
                )
                initial_mask = torch.from_numpy(initial_mask.astype(np.float32))

                # prepare color filter if needed
                from waveprop.devices import slm_dict
                from waveprop.devices import SLMParam as SLMParam_wp

                slm_param = slm_dict[config["trainable_mask"]["slm"]]
                if (
                    config["trainable_mask"]["train_color_filter"]
                    and SLMParam_wp.COLOR_FILTER in slm_param.keys()
                ):
                    color_filter = slm_param[SLMParam_wp.COLOR_FILTER]
                    color_filter = torch.from_numpy(color_filter.copy()).to(dtype=torch.float32)

                    # TODO: add small random values?
                    color_filter = color_filter + 0.1 * torch.rand_like(color_filter)

            else:
                raise ValueError(
                    f"Initial PSF value {config['trainable_mask']['initial_value']} not supported"
                )

            # convert to grayscale if needed
            if config["trainable_mask"]["grayscale"] and not is_grayscale(initial_mask):
                initial_mask = rgb2gray(initial_mask)

            mask = mask_class(
                initial_mask,
                downsample=downsample,
                color_filter=color_filter,
                **config["trainable_mask"],
            )

    return mask



import setuptools

# import version from file -> `__version__`
__version__ = None
with open("lensless/version.py") as f:
    exec(f.read())
assert __version__ is not None

# with open("README.rst", "r", encoding="utf-8") as fh:
#     long_description = fh.read()
long_description = "See the documentation at https://lensless.readthedocs.io/en/latest/"

setuptools.setup(
    name="lensless",
    version=__version__,
    author="Eric Bezzam",
    author_email="ebezzam@gmail.com",
    description="All-in-one package for lensless imaging: design, simulation, measurement, reconstruction.",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    url="https://github.com/LCAV/LenslessPiCam",
    packages=setuptools.find_namespace_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.8.1, <3.12",
    install_requires=[
        "opencv-python>=4.5.1.48",
        "numpy==1.26.4; python_version=='3.11'",
        "numpy>=1.22",
        "scipy>=1.7.0",
        "image>=1.5.33",
        "matplotlib>=3.4.2",
        "rawpy>=0.16.0",  # less than python 3.12
        "paramiko>=3.2.0",
        "hydra-core",
    ],
    extra_requires={"dev": ["pudb", "black"]},
)



"""

Prior to launching, launch `feh` image viewer from Raspberry Pi:
```
feh LenslessPiCam_display --scale-down --auto-zoom -R 0.1 -x -F -Y
```
Refresh shouldn't be too fast, otherwise copying may have an issue.

Consequently, after running this script, the displayed image will be reloaded
on the display of the Raspberry Pi.
```
python scripts/measure/remote_display.py
```

Check out the `configs/demo.yaml` file for parameters, specifically:

- `fp`: path to image to display
- `display`: parameters for displaying image

Procedure is as follows:
- Image is copied to Raspberry Pi
- On the Raspberry Pi it is padded accordingly and saved to the path being viewed
by `feh`

"""

import os
import numpy as np
from PIL import Image
import hydra
from lensless.hardware.utils import display
from lensless.hardware.utils import check_username_hostname


@hydra.main(version_base=None, config_path="../../configs", config_name="demo")
def remote_display(config):

    check_username_hostname(config.rpi.username, config.rpi.hostname)
    username = config.rpi.username
    hostname = config.rpi.hostname

    fp = config.fp
    shape = np.array(config.display.screen_res)
    psf = config.display.psf
    black = config.display.black
    white = config.display.white

    if psf:
        point_source = np.zeros(tuple(shape) + (3,))
        mid_point = shape // 2
        start_point = mid_point - psf // 2
        end_point = start_point + psf
        point_source[start_point[0] : end_point[0], start_point[1] : end_point[1]] = 255
        fp = "tmp_display.png"
        im = Image.fromarray(point_source.astype("uint8"), "RGB")
        im.save(fp)

    elif black:
        point_source = np.zeros(tuple(shape) + (3,))
        fp = "tmp_display.png"
        im = Image.fromarray(point_source.astype("uint8"), "RGB")
        im.save(fp)

    elif white:
        point_source = np.ones(tuple(shape) + (3,)) * 255
        fp = "tmp_display.png"
        im = Image.fromarray(point_source.astype("uint8"), "RGB")
        im.save(fp)

    """ processing on remote machine, less issues with copying """
    # copy picture to Raspberry Pi
    print("\nCopying over picture...")
    display(fp=fp, rpi_username=username, rpi_hostname=hostname, **config.display)

    if psf or black or white:
        os.remove(fp)


if __name__ == "__main__":
    remote_display()



from waveprop.dataset_util import SimulatedPytorchDataset
import hydra
from hydra.utils import to_absolute_path
import os
from lensless.utils.io import load_psf
from lensless.utils.image import rgb2gray
import torch
import time
import numpy as np
from torchvision import transforms, datasets
from tqdm import tqdm


@hydra.main(
    version_base=None, config_path="../../configs/simulate", config_name="simulate_torch_dataset"
)
def simulate(config):

    # load dataset
    transforms_list = [transforms.ToTensor()]
    if config.simulation.grayscale:
        transforms_list.append(transforms.Grayscale())
    transform = transforms.Compose(transforms_list)
    if config.files.dataset == "mnist":
        ds = datasets.MNIST(root="data", train=True, download=True, transform=transform)
    elif config.files.dataset == "fashion_mnist":
        ds = datasets.FashionMNIST(root="data", train=True, download=True, transform=transform)
    elif config.files.dataset == "cifar10":
        ds = datasets.CIFAR10(root="data", train=True, download=True, transform=transform)
    else:
        raise NotImplementedError(f"Dataset {config.files.dataset} not implemented.")

    # load PSF
    if config.files.psf is not None:
        psf_fp = to_absolute_path(config.files.psf)
        assert os.path.exists(psf_fp), f"PSF {psf_fp} does not exist."
        psf = load_psf(psf_fp, downsample=config.simulation.downsample)
        if config.simulation.grayscale:
            psf = rgb2gray(psf)
        psf = transforms.ToTensor()(psf[0])  # first depth
    else:
        assert config.simulation.output_dim is not None
        psf = None

    # batch_size = config.files.batch_size
    batch_size = config.files.batch_size
    n_files = config.files.n_files
    device_conv = config.device_conv
    target = config.target

    # check if gpu is available
    if device_conv == "cuda" and torch.cuda.is_available():
        print("Using GPU for convolution.")
        device_conv = "cuda"
    else:
        print("Using CPU for convolution.")
        device_conv = "cpu"

    # create Pytorch dataset and dataloader
    if n_files is not None:
        ds = torch.utils.data.Subset(ds, np.arange(n_files))
    ds_prop = SimulatedPytorchDataset(
        dataset=ds, psf=psf, device_conv=device_conv, target=target, **config.simulation
    )
    ds_loader = torch.utils.data.DataLoader(dataset=ds_prop, batch_size=batch_size, shuffle=True)

    # loop over batches
    start_time = time.time()
    for i, (x, target) in enumerate(tqdm(ds_loader)):

        if i == 0:
            print("Batch shape : ", x.shape)
            print("Target shape : ", target.shape)
            print("Batch device : ", x.device)

    print("Time per batch : ", (time.time() - start_time) / len(ds_loader))
    print(f"Went through {len(ds_loader)} batches.")


if __name__ == "__main__":
    simulate()
